# Copyright (c) 2014 Brocade Communications Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""NOS NETCONF XML Configuration Command Templates.

Interface Configuration Commands
"""

# Get NOS Version
SHOW_FIRMWARE_VERSION = (
    "show-firmware-version xmlns:nc="
    "'urn:brocade.com:mgmt:brocade-firmware-ext'"
)
GET_VCS_DETAILS = (
    'get-vcs-details xmlns:nc="urn:brocade.com:mgmt:brocade-vcs"'
)
SHOW_VIRTUAL_FABRIC = (
    'show-virtual-fabric xmlns:nc="urn:brocade.com:mgmt:brocade-vcs"'
)
GET_VIRTUAL_FABRIC_INFO = (
    'interface xmlns:nc="urn:brocade.com:mgmt:brocade-firmware-ext"'
)

NOS_VERSION = "./*/{urn:brocade.com:mgmt:brocade-firmware-ext}os-version"
VFAB_ENABLE = "./*/*/*/{urn:brocade.com:mgmt:brocade-vcs}vfab-enable"

# Create VLAN (vlan_id)
CREATE_VLAN_INTERFACE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <interface-vlan xmlns="urn:brocade.com:mgmt:brocade-interface">
            <interface>
                <vlan>
                    <name>{vlan_id}</name>
                </vlan>
            </interface>
        </interface-vlan>
    </config>
"""

# Delete VLAN (vlan_id)
DELETE_VLAN_INTERFACE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <interface-vlan xmlns="urn:brocade.com:mgmt:brocade-interface">
            <interface>
                <vlan operation="delete">
                    <name>{vlan_id}</name>
                </vlan>
            </interface>
        </interface-vlan>
    </config>
"""

#
# AMPP Life-cycle Management Configuration Commands
#

# Create AMPP port-profile (port_profile_name)
CREATE_PORT_PROFILE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <name>{name}</name>
        </port-profile>
    </config>
"""

# Create VLAN sub-profile for port-profile (port_profile_name)
CREATE_VLAN_PROFILE_FOR_PORT_PROFILE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <name>{name}</name>
            <vlan-profile/>
        </port-profile>
    </config>
"""

# Configure L2 mode for VLAN sub-profile (port_profile_name)
CONFIGURE_L2_MODE_FOR_VLAN_PROFILE_IN_DOMAIN = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <name>{name}</name>
            <vlan-profile>
                <switchport-basic>
                   <basic/>
                </switchport-basic>
            </vlan-profile>
        </port-profile>
    </config>
"""

# Configure L2 mode for VLAN sub-profile (port_profile_name)
CONFIGURE_L2_MODE_FOR_VLAN_PROFILE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <name>{name}</name>
            <vlan-profile>
                <switchport/>
            </vlan-profile>
        </port-profile>
    </config>
"""

# Configure trunk mode for VLAN sub-profile (port_profile_name)
CONFIGURE_TRUNK_MODE_FOR_VLAN_PROFILE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <name>{name}</name>
            <vlan-profile>
                <switchport>
                    <mode>
                        <vlan-mode>trunk</vlan-mode>
                    </mode>
                </switchport>
            </vlan-profile>
        </port-profile>
    </config>
"""

# Configure allowed VLANs for VLAN sub-profile
# (port_profile_name, allowed_vlan, native_vlan)
CONFIGURE_ALLOWED_VLANS_FOR_VLAN_PROFILE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <name>{name}</name>
            <vlan-profile>
                <switchport>
                    <trunk>
                        <allowed>
                            <vlan>
                                <add>{vlan_id}</add>
                            </vlan>
                        </allowed>
                    </trunk>
                </switchport>
            </vlan-profile>
        </port-profile>
    </config>
"""

# Delete port-profile (port_profile_name)
DELETE_PORT_PROFILE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile
xmlns="urn:brocade.com:mgmt:brocade-port-profile" operation="delete">
            <name>{name}</name>
        </port-profile>
    </config>
"""

# Activate port-profile (port_profile_name)
ACTIVATE_PORT_PROFILE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile-global xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <port-profile>
                <name>{name}</name>
                <activate/>
            </port-profile>
        </port-profile-global>
    </config>
"""

# Deactivate port-profile (port_profile_name)
DEACTIVATE_PORT_PROFILE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile-global xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <port-profile>
                <name>{name}</name>
                <activate
xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0" nc:operation="delete" />
            </port-profile>
        </port-profile-global>
    </config>
"""

# Associate MAC address to port-profile (port_profile_name, mac_address)
ASSOCIATE_MAC_TO_PORT_PROFILE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile-global xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <port-profile>
                <name>{name}</name>
                <static>
                    <mac-address>{mac_address}</mac-address>
                </static>
            </port-profile>
        </port-profile-global>
    </config>
"""

# Dissociate MAC address from port-profile (port_profile_name, mac_address)
DISSOCIATE_MAC_FROM_PORT_PROFILE = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile-global xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <port-profile>
                <name>{name}</name>
                <static
xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0" nc:operation="delete">
                    <mac-address>{mac_address}</mac-address>
                </static>
            </port-profile>
        </port-profile-global>
    </config>
"""

# port-profile domain management commands
REMOVE_PORTPROFILE_FROM_DOMAIN = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile-domain xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <port-profile-domain-name>{domain_name}</port-profile-domain-name>
                <profile  operation="delete">
                    <profile-name>{name}</profile-name>
                </profile>
            </port-profile-domain>
    </config>
"""
# put port profile in default domain
CONFIGURE_PORTPROFILE_IN_DOMAIN = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <port-profile-domain xmlns="urn:brocade.com:mgmt:brocade-port-profile">
            <port-profile-domain-name>{domain_name}</port-profile-domain-name>
                <profile>
                    <profile-name>{name}</profile-name>
                </profile>
            </port-profile-domain>
    </config>
"""

#
# L3 Life-cycle Management Configuration Commands
#

# Create SVI and assign ippaddres (rbridge_id,vlan_id,ip_address)
CONFIGURE_SVI_WITH_IP_ADDRESS = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
         <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
            <interface xmlns="urn:brocade.com:mgmt:brocade-interface">
                <ve>
                    <name>{vlan_id}</name>
                    <ip xmlns="urn:brocade.com:mgmt:brocade-ip-config">
                        <ip-config>
                          <address>
                             <address>{ip_address}</address>
                          </address>
                        </ip-config>
                    </ip>
                </ve>
            </interface>
         </rbridge-id>
    </config>
"""

# delete SVI (rbridge_id,vlan_id)
DELETE_SVI = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
         <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
            <interface xmlns="urn:brocade.com:mgmt:brocade-interface">
                <ve operation="delete">
                    <name>{vlan_id}</name>
                </ve>
            </interface>
         </rbridge-id>
    </config>
"""

# Activate SVI (rbridge_id,vlan_id)
ACTIVATE_SVI = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
         <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
            <interface xmlns="urn:brocade.com:mgmt:brocade-interface">
                <ve>
                    <name>{vlan_id}</name>
                    <shutdown xmlns="urn:brocade.com:mgmt:brocade-ip-config"
                    xc:operation="delete"></shutdown>
                </ve>
            </interface>
         </rbridge-id>
    </config>
"""

# Remove ipaddress from SVI (rbridge_id,vlan_id)
DECONFIGURE_IP_FROM_SVI = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
         <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
            <interface xmlns="urn:brocade.com:mgmt:brocade-interface">
                <ve>
                    <name>{vlan_id}</name>
                    <ip xmlns="urn:brocade.com:mgmt:brocade-ip-config">
                        <ip-config>
                            <address xc:operation="delete">
                                <address>{gw_ip}</address>
                            </address>
                        </ip-config>
                    </ip>
                </ve>
            </interface>
         </rbridge-id>
    </config>
"""

# create vrf (rbridge_id,vrf_name)
CREATE_VRF = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
         <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
                <vrf xmlns="urn:brocade.com:mgmt:brocade-vrf">
                    <vrf-name>{vrf_name}</vrf-name>
                </vrf>
         </rbridge-id>
    </config>
"""


# delete vrf (rbridge_id,vrf_name)
DELETE_VRF = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
        <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
                <vrf xmlns="urn:brocade.com:mgmt:brocade-vrf"
                       xc:operation="delete">
                    <vrf-name>{vrf_name}</vrf-name>
                </vrf>
         </rbridge-id>
    </config>
"""

# configure route distinguisher for vrf (rbridge_id,vrf_name, rd)
CONFIGURE_RD_FOR_VRF = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
         <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
            <vrf xmlns="urn:brocade.com:mgmt:brocade-vrf">
                <vrf-name>{vrf_name}</vrf-name>
                <route-distiniguisher>{rd}</route-distiniguisher>
            </vrf>
         </rbridge-id>
    </config>
"""

# configure address-family for vrf (rbridge_id,vrf_name)
ADD_ADDRESS_FAMILY_FOR_VRF_V1 = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
         <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
            <vrf xmlns="urn:brocade.com:mgmt:brocade-vrf">
                <vrf-name>{vrf_name}</vrf-name>
                <address-family xmlns="urn:brocade.com:mgmt:brocade-vrf">
                    <ipv4>
                        <max-route>1200</max-route>
                    </ipv4>
                </address-family>
            </vrf>
         </rbridge-id>
    </config>
"""

# configure address-family for vrf (rbridge_id,vrf_name)
ADD_ADDRESS_FAMILY_FOR_VRF = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
         <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
            <vrf xmlns="urn:brocade.com:mgmt:brocade-vrf">
                <vrf-name>{vrf_name}</vrf-name>
                <address-family xmlns="urn:brocade.com:mgmt:brocade-vrf">
                    <ip>
                        <unicast/>
                    </ip>
                </address-family>
            </vrf>
         </rbridge-id>
    </config>
"""

# Bind vrf to SVI (rbridge_id,vlan_idi, vrf)
ADD_VRF_TO_SVI = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
         <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
            <interface xmlns="urn:brocade.com:mgmt:brocade-interface">
                <ve>
                    <name>{vlan_id}</name>
                    <vrf xmlns="urn:brocade.com:mgmt:brocade-ip-config">
                        <forwarding>{vrf_name}</forwarding>
                    </vrf>
                </ve>
            </interface>
         </rbridge-id>
    </config>
"""

# unbind  vrf from SVI (rbridge_id,vlan_idi, vrf)
DELETE_VRF_FROM_SVI = """
    <config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
         <rbridge-id xmlns="urn:brocade.com:mgmt:brocade-rbridge">
            <rbridge-id>{rbridge_id}</rbridge-id>
            <interface xmlns="urn:brocade.com:mgmt:brocade-interface">
                <ve>
                    <name>{vlan_id}</name>
                    <vrf xmlns="urn:brocade.com:mgmt:brocade-ip-config"
                            operation="delete">
                        <forwarding>{vrf_name}</forwarding>
                    </vrf>
                </ve>
            </interface>
         </rbridge-id>
    </config>
"""

#
# Constants
#

# Port profile naming convention for Neutron networks
OS_PORT_PROFILE_NAME = "openstack-profile-{id}"
OS_VRF_NAME = "osv-{id}"

# Port profile filter expressions
PORT_PROFILE_XPATH_FILTER = "/port-profile"
PORT_PROFILE_NAME_XPATH_FILTER = "/port-profile[name='{name}']"



# Copyright 2015 Brocade Communications System, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""Implementation of Brocade L3RouterPlugin for NI devices."""

from neutron.i18n import _LE
from neutron.i18n import _LI
from neutron.plugins.ml2.driver_context import NetworkContext
from neutron.services.l3_router import l3_router_plugin as router
from oslo_log import log as logging
from oslo_utils import importutils

LOG = logging.getLogger(__name__)
NI = "networking_brocade.mlx.ml2.fi_ni.ni_driver.NetIronDriver"
BROCADE_CONFIG = ("networking_brocade.mlx.ml2.fi_ni.brcd_config."
                  "ML2BrocadeConfig")
DRIVER_FACTORY = ("networking_brocade.mlx.ml2.fi_ni."
                  "driver_factory.BrocadeDriverFactory")
# Property to identify which device type to use
# Currently supports only NI devices. If FI devices needs to be
# supported add FI to the list as shown below:
# ['NI', 'FI']
ROUTER_DEVICE_TYPE = ['NI']


class BrocadeRouterPlugin(router.L3RouterPlugin):

    """
    SVI Mechanism driver for Brocade ICX and MLX devices. This will take care
    of Create/Delete router interface to introduce L3 support.
    """

    def __init__(self):
        """Initialize Brocade Plugin

        Specify switch address and db configuration.
        """
        self._devices = {}
        self._physical_networks = {}
        self._driver_map = {}
        super(BrocadeRouterPlugin, self).__init__()
        self.brocade_init()

    def brocade_init(self):
        """Brocade specific initialization. Parses the configuration template
        and stores all the device information in a dictionary. Then it filters
        the router devices from it.
        """
        LOG.debug("BrocadeRouterPlugin::brocade_init()")

        self._devices, self._physical_networks = importutils.import_object(
            BROCADE_CONFIG).create_brocade_dictionary(isL2=False)

    def add_router_interface(self, context, router_id, interface_info):
        """Adds router interface to a vlan on NI device and assigns ip
         address to configure l3 router interface.
        """

        return self.add_remove_router_interface(context, router_id,
                                         interface_info,
                                         True)

    def remove_router_interface(self, context, router_id, interface_info):
        """Removes router interface on NI device
        """
        return self.add_remove_router_interface(context, router_id,
                                                interface_info,
                                                False)

    def add_remove_router_interface(self, context, router_id, interface_info,
                                    is_add):
        """
        Add/Remove router interface on NI device

        :param:context: Contains the network details
        :param:router_id: The router ID
        :param:interface_info: Contains interface details
        :param:is_add: True for add operation, False for remove operation.
        :raises: Exception
        """
        operation = None
        method = None
        if is_add is True:
            method = "add_router_interface"
            operation = "Add"
        else:
            method = "remove_router_interface"
            operation = "Remove"
        info = getattr(super(BrocadeRouterPlugin, self),
                       method)(context, router_id, interface_info)
        interface_info = info
        subnet = self._core_plugin._get_subnet(context.elevated(),
                                               interface_info["subnet_id"])
        vlan_id, gateway_ip_cidr = self._get_network_info(context, subnet)
        for device in self._devices:
            device_info = self._devices.get(device)
            address = device_info.get('address')
            driver = self._get_driver(device)
            LOG.info(_LI("BrocadeRouterPlugin:Before %(op)s l3 "
                         "router to vlan %(vlan_id)s with ip %(gatewayip)s"
                         "on device "
                         "%(host)s"), {'op': operation,
                                       'vlan_id': vlan_id,
                                       'gatewayip': gateway_ip_cidr,
                                       'host': address})
            try:
                if is_add is True:
                    getattr(driver, method)(vlan_id, gateway_ip_cidr)
                else:
                    getattr(driver, method)(vlan_id)
            except Exception as e:
                if is_add:
                    method = "remove_router_interface"
                    info = getattr(super(BrocadeRouterPlugin, self),
                                   method)(context, router_id, interface_info)
                LOG.exception(_LE("BrocadeRouterPlugin: failed to"
                                  " %(op)s l3 router interface for "
                                  "device= %(device)s exception : "
                                  "%(error)s"), {'op': operation,
                                                 'device': address,
                                                 'error': e.args})
                raise Exception(
                    _("BrocadeRouterPlugin: %(op)s router "
                      "interface failed"), {'op': operation})
            LOG.info(_LI("BrocadeRouterPlugin:%(op)sed router "
                         "interface in vlan = %(vlan_id)s with ip address"
                         " %(gatewayip)s on device %(host)s "
                         "successful"), {'op': operation,
                                         'vlan_id': vlan_id,
                                         'gatewayip': gateway_ip_cidr,
                                         'host': address})
        return info

    def _get_driver(self, device):
        """
        Gets the driver based on the firmware version of the device.

        :param:device: Contains device name
        :raises: Exception
        """
        driver = self._driver_map.get(device)
        if driver is None:
            driver_factory = importutils.import_object(DRIVER_FACTORY)
            device_info = self._devices.get(device)
            address = device_info.get('address')
            os_type = device_info.get('ostype')
            try:
                driver = driver_factory.get_driver(device_info)
                if driver is not None and os_type in ROUTER_DEVICE_TYPE:
                    self._driver_map.update({device: driver})
            except Exception as e:
                LOG.exception(_LE("BrocadeRouterPlugin:"
                                  "_filter_router_devices: Error while getting"
                                  " driver : device - %(host)s: %(error)s"),
                              {'host': address, 'error': e.args})
                raise Exception(_("BrocadeRouterPlugin:"
                                  "_filter_router_devices failed for "
                                  "device %(host)s"), {'host': address})
        return driver

    def _get_network_info(self, context, subnet):
        """
        Gets the network info from the context and brocade db

        :param:context: Contains the network details
        :param:subnet: Contains the subnet details
        :returns:vlan_id - VLAN ID of the network
        :returns:gateway_ip_cidr - Gateway IP address with network length.
            Example: 1.1.1.1/24
        """
        cidr = subnet["cidr"]
        net_addr, net_len = self.net_addr(cidr)
        gateway_ip = subnet["gateway_ip"]
        network_id = subnet['network_id']
        ml2_db = NetworkContext(self, context, {'id': network_id})
        vlan_id = ml2_db.network_segments[0]['segmentation_id']
        gateway_ip_cidr = gateway_ip + '/' + str(net_len)

        return vlan_id, gateway_ip_cidr

    @staticmethod
    def net_addr(addr):
        """Get network address prefix and length from a given address."""
        if addr is None:
            return None, None
        nw_addr, nw_len = addr.split('/')
        nw_len = int(nw_len)
        return nw_addr, nw_len



# Copyright 2015 Brocade Communications System, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Driver to interact with fastiron devices. Currently has no specific
implementations. Refer EthernetDriver
"""

from networking_brocade.mlx.ml2.fi_ni.fi_ni_driver import FiNiDriver


class FastIronDriver(FiNiDriver):

    """
    Driver to interact with fastiron devices. Currently has no specific
    implementations. Refer EthernetDriver
    """

    def __init__(self, device):
        super(FastIronDriver, self).__init__(device)

    def create_l3_router(self, vlan_id, gateway_ip_cidr):
        """
        Create router interface.

        :param:vlan_id: vlan id to which the created router interface will
            be part of
        :param:gateway_ip_cidr: Gateway IP address with network length.
        :raises: Exception
        """
        raise Exception("create_l3_router - This operation is currently not"
                        " supported in this platform ")

    def delete_l3_router(self, vlan_id):
        """
        Remove router interface.

        :param:vlan_id: vlan id to which the router interface is part of
        :raises: Exception
        """
        raise Exception("delete_l3_router - This operation is currently not"
                        " supported in this platform ")



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import requests
import urllib

from neutron import context
from neutron.db import models_v2
from neutron.openstack.common import uuidutils
from neutron.tests import base as n_base

from networking_brocade.vyatta.common import utils as vyatta_utils
from networking_brocade.vyatta.vrouter import client as vyatta_client
from networking_brocade.vyatta.vrouter import driver as vrouter_driver


_uuid = uuidutils.generate_uuid


def mock_patch(target, new=mock.DEFAULT):
    patcher = mock.patch(target, new)
    return patcher.start()


def mock_object(target, attribute, new=mock.DEFAULT):
    patcher = mock.patch.object(target, attribute, new)
    return patcher.start()


class TestVRouterDriver(n_base.BaseTestCase):
    def setUp(self):
        super(TestVRouterDriver, self).setUp()

        # mock_object(
        #     vrouter_driver.VyattaVRouterDriver,
        #     '_management_network').return_value = _uuid()

        self._nova_client = mock_patch(
            'novaclient.v1_1.client.Client').return_value

        self._server = self._create_server()
        self._nova_client.servers.create.return_value = self._server
        self._nova_client.servers.get.return_value = self._server

        self._router_api_mock = mock_object(
            vrouter_driver.VyattaVRouterDriver, '_get_router_api')
        self._router_api = self._router_api_mock.return_value

        self.driver = vrouter_driver.VyattaVRouterDriver()

    @staticmethod
    def _create_server():
        server = mock.Mock()
        server.id = _uuid()
        server.status = 'ACTIVE'
        server.interface_list.return_value = ['mng-iface']

        return server

    def test_create_router(self):
        ctx = context.Context('', 'tenant_id1')
        router = self.driver.create_router(ctx)
        self.assertEqual(self._server.id, router)

    def test_init_router(self):
        ctx = context.Context('', 'tenant_id1')

        router = {'id': _uuid()}

        self.driver.init_router(ctx, router)

        self._router_api_mock.assert_called_once_with(mock.ANY, router['id'])
        self._router_api.init_router.assert_called_once_with(
            mock.ANY, mock.ANY)

    def test_delete_router(self):
        ctx = context.Context('', 'tenant_id1')

        router_id = _uuid()

        self.driver.delete_router(ctx, router_id)

        self._router_api_mock.assert_called_once_with(mock.ANY, router_id)
        self._router_api.disconnect.assert_called_once_with()
        self._nova_client.servers.delete.assert_called_once_with(router_id)

    def test_attach_interface(self):
        ctx = context.Context('', 'tenant_id1')
        port_id = _uuid()

        self.driver.attach_interface(ctx, self._server.id, port_id)

        self._nova_client.servers.get.assert_called_once_with(self._server.id)
        self._server.interface_attach.assert_called_once_with(
            port_id, mock.ANY, mock.ANY)

    def test_detach_interface(self):
        ctx = context.Context('', 'tenant_id1')
        port_id = _uuid()

        self.driver.detach_interface(ctx, self._server.id, port_id)

        self._nova_client.servers.get.assert_called_once_with(self._server.id)
        self._server.interface_detach.assert_called_once_with(port_id)

    def test_configure_interface(self):

        ctx = context.Context('', 'tenant_id1')
        router_id = _uuid()
        interfaces = ['eth0', 'eth1']

        self.driver.configure_interface(ctx, router_id, interfaces)

        for interface in interfaces:
            self._router_api.add_interface_to_router.assert_any_call(interface)

    def test_deconfigure_interface(self):
        ctx = context.Context('', 'tenant_id1')
        router_id = _uuid()
        interfaces = ['eth0', 'eth1']

        self.driver.deconfigure_interface(ctx, router_id, interfaces)

        for interface in interfaces:
            self._router_api.remove_interface_from_router.assert_any_call(
                interface)

    def test_configure_gateway(self):
        ctx = context.Context('', 'tenant_id1')
        router_id = _uuid()
        interfaces = ['eth0']

        self.driver.configure_gateway(ctx, router_id, interfaces)

        self._router_api.update_router.assert_called_once_with(
            external_gateway_info=interfaces[0])

    def test_clear_gateway(self):
        ctx = context.Context('', 'tenant_id1')
        router_id = _uuid()

        self.driver.clear_gateway(ctx, router_id)

        self._router_api.update_router.assert_called_once_with(
            external_gateway_info=None)

    def test_assign_floating_ip(self):
        ctx = context.Context('', 'tenant_id1')
        router_id = _uuid()
        floating_ip = '192.168.1.13'
        fixed_ip = '10.10.1.13'

        self.driver.assign_floating_ip(ctx, router_id, floating_ip, fixed_ip)

        self._router_api.assign_floating_ip.assert_called_once_with(
            floating_ip, fixed_ip)

    def test_unassign_floating_ip(self):
        ctx = context.Context('', 'tenant_id1')
        router_id = _uuid()
        floating_ip = '192.168.1.13'
        fixed_ip = '10.10.1.13'

        self.driver.unassign_floating_ip(ctx, router_id, floating_ip, fixed_ip)

        self._router_api.unassign_floating_ip.assert_called_once_with(
            floating_ip, fixed_ip)

    def test_update_static_routes(self):
        ctx = context.Context('', 'tenant_idl')
        router_id = _uuid()

        RouteRule = vyatta_utils.RouteRule
        routes_to_add = (
            RouteRule(dest_cidr='10.1.0.0/24', next_hop='192.168.1.1'),
            RouteRule(dest_cidr='10.2.0.0/24', next_hop='192.168.1.1')
        )
        routes_to_del = (
            RouteRule(dest_cidr='10.3.0.0/24', next_hop='192.168.1.1'),
        )

        self.driver.update_static_routes(ctx, router_id,
                                         routes_to_add, routes_to_del)
        self._router_api.update_static_routes.assert_called_once_with(
            routes_to_add, routes_to_del)


class TestVRouterDriverApi(n_base.BaseTestCase):

    def setUp(self):
        super(TestVRouterDriverApi, self).setUp()

        self._nova_client = mock_patch(
            'novaclient.v1_1.client.Client').return_value

        # self._mock_object(vrouter_driver.VyattaVRouterDriver,
        #                   '_management_network')

        self.driver = vrouter_driver.VyattaVRouterDriver()

    def _mock_object(self, target, attribute, new=mock.DEFAULT):
        patcher = mock.patch.object(target, attribute, new)
        return patcher.start()

    def test_get_router_api(self):
        ctx = mock.Mock()
        router_id = _uuid()

        router_addr = '172.16.18.10'

        server = self._nova_client.servers.get.return_value
        server.addresses = {
            'external': [
                {'addr': router_addr, }
            ]
        }

        query = ctx.session.query.return_value
        query.filter_by.return_value.one.return_value = models_v2.Network(
            name='external')

        with mock.patch.object(
                vyatta_client.VRouterRestAPIClient, 'connect') as api_connect:
            api = self.driver._get_router_api(ctx, router_id)
            self.assertIsNotNone(api)
            api_connect.assert_called_once_with(router_addr)


SHOW_VERSION_OUTPUT = """
Version:      VSE6.6R5X3
Description:  Brocade Vyatta 5410 vRouter 6.6 R5X3
Copyright:    2006-2014 Vyatta, Inc.
"""

SHOW_CONFIG_OUTPUT = """
interfaces {
    ethernet eth0 {
        address 10.10.0.1
        description External_Gateway
        hw-id 08:00:27:02:b4:67
        duplex auto
        smp_affinity auto
        speed auto
    }
    loopback lo {
    }
}
service {
    https {
        http-redirect disable
    }
    lldp {
        interface all {
        }
        snmp {
            enable
        }
    }
    snmp {
        community public {
            authorization ro
        }
    }
    ssh {
        port 22
    }
}
system {
    host-name router1
    login {
        user vyatta {
            authentication {
                encrypted-password ****************
            }
            level admin
        }
    }
    syslog {
        global {
            facility all {
                level debug
            }
            facility protocols {
                level debug
            }
        }
        user all {
            facility all {
                level emerg
            }
        }
    }
    time-zone GMT
}
"""


class VRouterRestAPIClientMixin(object):

    def _create_client(self):

        self.address = 'example.com'

        client = vyatta_client.VRouterRestAPIClient()
        client.address = self.address
        client._vrouter_model = (
            vyatta_client.VRouterRestAPIClient._VROUTER_VR_MODEL)

        return client


class TestVRouterRestAPIClient(n_base.BaseTestCase,
                               VRouterRestAPIClientMixin):

    def setUp(self):
        super(TestVRouterRestAPIClient, self).setUp()

        self._rest_mock = mock_object(
            vyatta_client.VRouterRestAPIClient, '_rest_call')
        self._rest_mock.side_effect = self._rest_call

        m = mock_object(vyatta_client.VRouterRestAPIClient,
                        'get_ethernet_if_id')
        m.return_value = 'eth0'

        m = mock_object(
            vyatta_client.VRouterRestAPIClient, '_get_admin_state')
        m.return_value = False

    def _check_cmd(self, cmd, check_type=None, check_cmd=None):
        if check_type is not None:
            self.assertEqual(cmd.cmd_type, check_type)
        if check_cmd is not None:
            self.assertEqual(cmd.cmd, check_cmd)

    @staticmethod
    def _rest_call(method, uri, headers=None, session=None):
        response = mock.Mock()
        response.status_code = requests.codes.OK
        response.reason = 'Ok'
        response.text = '{}'
        response.headers = {
            'Location': 'location'
        }

        return response

    def _check_update_interface(self, action, interface_info):
        prefix = '/location/{action}/interfaces/dataplane/eth0'.format(
            action=action)

        self._rest_mock.assert_any_call(
            'PUT', prefix + '/address/' + interface_info['ip_address'])
        self._rest_mock.assert_any_call(
            'POST', '/location/commit')
        self._rest_mock.assert_any_call(
            'POST', '/location/save')

    def test_init_router(self):
        client = self._create_client()
        with mock.patch('eventlet.greenthread.sleep'):
            client.init_router('router1', True)

        self._rest_mock.assert_any_call(
            'PUT', '/location/delete/system/ip/disable-forwarding',
        )

    def test_add_interface_to_router(self):

        interface_info = {
            'mac_address': '08:00:27:02:b4:67',
            'ip_address': '10.10.16.20',
            'gateway_ip': '10.10.16.1'
        }

        client = self._create_client()
        client.add_interface_to_router(interface_info)

        self._check_update_interface('set', interface_info)

    def test_remove_interface_from_router(self):

        interface_info = {
            'mac_address': '08:00:27:02:b4:67',
            'ip_address': '10.10.16.20',
            'gateway_ip': '10.10.16.1'
        }

        client = self._create_client()
        client.remove_interface_from_router(interface_info)

        self._check_update_interface('delete', interface_info)

    def test_assign_floating_ip(self):
        client = self._create_client()
        client._external_gw_info = vyatta_client.InterfaceInfo(
            'eth0', '172.16.18.10', '172.16.18.1')
        client.assign_floating_ip('172.16.18.12', '10.10.10.12')

        self.assertIn('172.16.18.12.10.10.10.12', client._floating_ip_dict)

        client.unassign_floating_ip('172.16.18.12', '10.10.10.12')

        self.assertNotIn('172.16.18.12.10.10.10.12', client._floating_ip_dict)

    def test_update_static_routes(self):
        cmd_batch_mock = mock.Mock()
        mock_object(vyatta_client.VRouterRestAPIClient,
                    'exec_cmd_batch', cmd_batch_mock)

        RouteRule = vyatta_utils.RouteRule

        routes_to_add = tuple((
            RouteRule(dest_cidr='10.1.0.0/24', next_hop='192.168.1.1'),
            RouteRule(dest_cidr='10.2.0.0/24', next_hop='192.168.1.1'),
        ))
        routes_to_del = tuple((
            RouteRule(dest_cidr='10.3.0.0/24', next_hop='192.168.1.1'),
        ))

        client = self._create_client()
        client.update_static_routes(routes_to_add, routes_to_del)

        expected_batch = list()
        for rule in routes_to_add:
            cmd = vyatta_client.SetCmd((
                'protocols/static/route/{0}/next-hop/{1}').format(
                    urllib.quote_plus(rule.dest_cidr),
                    urllib.quote_plus(rule.next_hop)))
            expected_batch.append(cmd)
        for rule in routes_to_del:
            cmd = vyatta_client.DeleteCmd('protocols/static/route/{0}'.format(
                urllib.quote_plus(rule.dest_cidr)))
            expected_batch.append(cmd)

        cmd_batch_mock.assert_called_once_with(expected_batch)

    def test_update_router(self):
        client = self._create_client()

        gw_info = {
            'mac_address': '08:00:27:02:b4:67',
            'ip_address': '10.10.16.20',
            'gateway_ip': '10.10.16.1'
        }

        client.update_router('rotuer1', True, gw_info)

        self._check_update_interface('set', gw_info)
        self._rest_mock.assert_any_call(
            'PUT', '/location/set/protocols/static/route/0.0.0.0%2F0/next-hop/'
            + gw_info['gateway_ip'])

        self._rest_mock.reset_mock()

        client.update_router('rotuer1', True)

        self._check_update_interface('delete', gw_info)
        self._rest_mock.assert_any_call(
            'PUT', '/location/delete/protocols/static/route/0.0.0.0%2F0')

    def test_get_nat_cmd(self):
        client = self._create_client()

        nat_cmd = client._get_nat_cmd()
        self.assertEqual(nat_cmd, 'service/nat')

        client._vrouter_model = (
            vyatta_client.VRouterRestAPIClient._VROUTER_VSE_MODEL)
        nat_cmd = client._get_nat_cmd()
        self.assertEqual(nat_cmd, 'nat')

    def test_add_snat_rule_cmd(self):

        client = self._create_client()
        cmd_list = []
        client._add_snat_rule_cmd(
            cmd_list, 2, 'eth1', '10.10.12.1', '192.168.12.1')

        prefix = 'service/nat/source/rule/2'

        self._check_cmd(cmd_list[0], 'set', prefix)
        self._check_cmd(cmd_list[1], 'set',
                        prefix + '/outbound-interface/eth1')
        self._check_cmd(cmd_list[2], 'set',
                        prefix + '/source/address/10.10.12.1')
        self._check_cmd(cmd_list[3], 'set',
                        prefix + '/translation/address/192.168.12.1')

    def test_add_dnat_rule_cmd(self):

        client = self._create_client()
        cmd_list = []
        client._add_dnat_rule_cmd(
            cmd_list, 2, 'eth1', '10.10.12.1', '192.168.12.1')

        prefix = 'service/nat/destination/rule/2'

        self._check_cmd(cmd_list[0], 'set', prefix)
        self._check_cmd(cmd_list[1], 'set', prefix + '/inbound-interface/eth1')
        self._check_cmd(cmd_list[2], 'set',
                        prefix + '/destination/address/10.10.12.1')
        self._check_cmd(cmd_list[3], 'set',
                        prefix + '/translation/address/192.168.12.1')

    def test_delete_snat_rule_cmd(self):
        client = self._create_client()
        cmd_list = []
        client._delete_snat_rule_cmd(cmd_list, 2)

        self._check_cmd(cmd_list[0], 'delete', 'service/nat/source/rule/2')

    def test_delete_dnat_rule_cmd(self):
        client = self._create_client()
        cmd_list = []
        client._delete_dnat_rule_cmd(cmd_list, 2)

        self._check_cmd(cmd_list[0],
                        'delete', 'service/nat/destination/rule/2')

    def test_set_router_name_cmd(self):
        client = self._create_client()
        cmd_list = []
        client._set_router_name_cmd(cmd_list, 'router1')

        self._check_cmd(cmd_list[0], 'set', 'system/host-name/router1')

    def test_set_system_gateway_cmd(self):
        client = self._create_client()
        cmd_list = []
        client._set_system_gateway_cmd(cmd_list, '10.10.16.1')

        self.assertEqual(len(cmd_list), 1)
        self._check_cmd(
            cmd_list[0], 'set',
            'protocols/static/route/0.0.0.0%2F0/next-hop/10.10.16.1')

    def test_delete_system_gateway_cmd(self):
        client = self._create_client()
        cmd_list = []
        # NOTE: Gateway parameter is not used
        client._delete_system_gateway_cmd(cmd_list, '10.10.16.1')

        self.assertEqual(len(cmd_list), 1)
        self._check_cmd(
            cmd_list[0], 'delete',
            'protocols/static/route/0.0.0.0%2F0')

    def test_configure_cmd_batch(self):
        client = self._create_client()
        cmd_list = [
            vyatta_client.SetCmd('cmd1'),
            vyatta_client.DeleteCmd('cmd2')
        ]

        client.exec_cmd_batch(cmd_list)

        self.assertEqual(
            self._rest_mock.call_count, len(cmd_list) + 4)

    def test_get_config_cmd(self):
        client = self._create_client()
        client._get_config_cmd('system/ip/disable-forwarding')

        self.assertEqual(self._rest_mock.call_count, 3)

    def test_show_cmd(self):

        self._rest_mock.side_effect = [
            self._make_http_response(201, headers={'Location': '/fake-url'}),
            self._make_http_response(200, text=SHOW_VERSION_OUTPUT),
            self._make_http_response(410),
            self._make_http_response(200)]

        client = self._create_client()
        client._show_cmd('version')
        self.assertEqual(self._rest_mock.call_count, 4)

    def test_process_model(self):
        client = vyatta_client.VRouterRestAPIClient()

        with mock.patch.object(
                vyatta_client.VRouterRestAPIClient, '_show_cmd') as show_cmd:
            show_cmd.return_value = SHOW_VERSION_OUTPUT
            client._process_model()

            self.assertEqual(client._vrouter_model, 54)

    def test_sync_cache(self):

        client = self._create_client()

        with mock.patch.object(
                vyatta_client.VRouterRestAPIClient, '_show_cmd') as show_cmd:
            show_cmd.return_value = SHOW_CONFIG_OUTPUT

            client._sync_cache()

        interface_info = vyatta_client.InterfaceInfo(
            'eth0', '10.10.0.1')
        self.assertEqual(client._external_gw_info, interface_info)

    def _make_http_response(self, status_code, headers=None, text=None):

        if headers is None:
            headers = {}

        response = mock.Mock()
        response.status_code = status_code
        response.headers = headers
        response.text = text

        return response


class TestLowLevelRestAPIClient(n_base.BaseTestCase,
                                VRouterRestAPIClientMixin):
    def test_get_admin_state(self):

        client = self._create_client()
        with mock.patch.object(
                vyatta_client.VRouterRestAPIClient,
                '_show_cmd') as get_cmd:
            get_cmd.return_value = "IP forwarding is on"

            state = client._get_admin_state()
            self.assertTrue(state)

    def test_get_ethernet_if_id(self):

        client = self._create_client()
        with mock.patch.object(
                vyatta_client.VRouterRestAPIClient,
                '_get_interfaces') as get_ifs:
            get_ifs.return_value = [{'name': 'dp0e5',
                                     'ip_addrs': '192.168.21.3',
                                     'mac_address': '08:00:27:4a:be:12'}]
            if_id = client.get_ethernet_if_id('08:00:27:4a:be:12')
            self.assertEqual(if_id, 'dp0e5')

    def test_rest_call(self):
        action = 'GET'
        uri = '/show/config'

        client = self._create_client()
        with mock.patch('requests.request') as request:
            client._rest_call(action, uri)

            uri = 'https://{0}{1}'.format(self.address, uri)

            request.assert_called_once_with(
                action, uri, verify=False, auth=mock.ANY,
                headers={'Content-Length': 0, 'Accept': 'application/json'})



# Copyright 2015 Brocade Communications Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from networking_brocade.mlx.ml2.fi_ni import (
    mechanism_brocade_fi_ni as brocadefinimechanism)
from neutron import context
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.tests import base
from oslo_log import log as logging
from oslo_utils import importutils

LOG = logging.getLogger(__name__)
MECHANISM_NAME = ('networking_brocade.mlx.ml2.fi_ni.'
                  'mechanism_brocade_fi_ni.BrocadeFiNiMechanism')
FAKE_PHYSICAL_NETWORK = 'physnet1'

config_map = {}

devices = {
    'icx': {
        'address': '1.1.1.1',
        'username': 'admin',
        'password': 'pass',
        'physical_networks': 'physnet1',
        'ports': '1/1/1,1/1/2',
        'os_type': 'FI'
    },
    'mlx': {
        'address': '2.2.2.2',
        'username': 'admin',
        'password': 'pass',
        'physical_networks': 'physnet1',
        'ports': '2/1,2/2',
        'os_type': 'NI'
    },
}

physical_networks = {
    'physnet1': ['icx', 'mlx']
}

actual_config_map = {
    '1.1.1.1': {
        200: ['1/1/1', '1/1/2']
    },
    '2.2.2.2': {
        200: ['2/1', '2/2']
    }
}

empty_config_map = {
}


class TestBrocadeFiNiMechDriver(base.BaseTestCase):

    """
    Test Brocade FI/NI mechanism driver.
    """

    def setUp(self):
        _mechanism_name = MECHANISM_NAME

        def mocked_initialize(self):
            self._devices = devices
            self._physical_networks = physical_networks

        with mock.patch.object(brocadefinimechanism.BrocadeFiNiMechanism,
                        'initialize', new=mocked_initialize):
            super(TestBrocadeFiNiMechDriver, self).setUp()
            self.mechanism_driver = importutils.import_object(_mechanism_name)

    def test_create_network_postcommit_wrong_physnet(self):
        """
        Test create network with wrong value for physical network.
        Physical network to which the devices belong is 'physnet1' but
        we make a call to create network with physical network 'physnet2'.
        In this case we raise an exception with error message -
        "Brocade Mechanism: failed to create network, network cannot be
        created in the configured physical network."
        """
        ctx = self._get_network_context('physnet2', 'vlan')
        self.assertRaises(ml2_exc.MechanismDriverError,
                          self.mechanism_driver.create_network_postcommit, ctx)

    def test_create_network_postcommit_wrong_network_type(self):
        """
        Test create network with wrong value for network type. The plugin
        allows to create network only if the request is to create a VLAN
        network. For any other network type following exception is raised -
        'Brocade Mechanism failed to create network, only network type vlan
        is supported"
        """
        ctx = self._get_network_context('physnet1', 'vxlan')
        self.assertRaises(ml2_exc.MechanismDriverError,
                          self.mechanism_driver.create_network_postcommit, ctx)

    @mock.patch.object(brocadefinimechanism.BrocadeFiNiMechanism,
                       '_get_driver')
    def test_create_network_postcommit(self, mock_driver):
        """
        Test create network with correct input values
        """
        mock_driver.side_effect = self.side_effect
        ctx = self._get_network_context('physnet1', 'vlan')
        self.mechanism_driver.create_network_postcommit(ctx)
        self.assertDictSupersetOf(config_map, actual_config_map)

    @mock.patch.object(brocadefinimechanism.BrocadeFiNiMechanism,
                       '_get_driver')
    def test_create_network_postcommit_driver_exception(self, mock_driver):
        """
        Expect an exception while trying to get the driver
        """
        mock_driver.side_effect = self.side_effect_error
        ctx = self._get_network_context('physnet1', 'vlan')
        self.assertRaises(ml2_exc.MechanismDriverError,
                          self.mechanism_driver.create_network_postcommit, ctx)

    @mock.patch.object(brocadefinimechanism.BrocadeFiNiMechanism,
                       '_get_driver')
    def test_create_network_postcommit_exception(self, mock_driver):
        """
        Exception is raised when the input values are incorrect.
        """
        mock_driver.side_effect = self.side_effect_error
        ctx = self._get_network_context('physnet1', 'vlan')
        self.assertRaises(ml2_exc.MechanismDriverError,
                          self.mechanism_driver.create_network_postcommit, ctx)

    @mock.patch.object(brocadefinimechanism.BrocadeFiNiMechanism,
                       '_get_driver')
    def test_delete_network_postcommit(self, mock_driver):
        """
        Test delete network
        """
        mock_driver.side_effect = self.side_effect
        ctx = self._get_network_context('physnet1', 'vlan')
        vlan_map = {200: ['1/1/1', '1/1/2']}
        config_map.update({'1.1.1.1': vlan_map})
        vlan_map = {200: ['2/1', '2/2']}
        config_map.update({'2.2.2.2': vlan_map})
        self.mechanism_driver.delete_network_postcommit(ctx)
        self.assertDictSupersetOf(config_map, empty_config_map)

    def _get_network_context(self, physnet, network_type):
        """
        Create mock network context
        """
        network = {
            'id': 1,
            'name': 'private',
            'tenant_id': 1,
            'vlan': 200,
            'network_type': network_type,
            'provider:segmentation_id': 200
        }

        network_segments = [{
            'id': 1,
            'segmentation_id': 200,
            'network_type': network_type,
            'physical_network': physnet
        }]

        _plugin_context = context.get_admin_context()
        return FakeNetworkContext(network, network_segments, _plugin_context)

    def side_effect(self, dev_name):
        """
        Mock _get_driver method and return FakeDriver
        """
        device = devices.get(dev_name)
        return FakeDriver(device)

    def side_effect_error(self, dev_name):
        """
        Mock _get_driver method and return FakeDriver
        """
        device = devices.get(dev_name)
        return FakeDriver(device, error=True)


class FakeNetworkContext(object):

    """To generate network context for testing purposes only."""

    def __init__(self, network, segments=None, original_network=None):
        self._network = network
        self._original_network = original_network
        self._segments = segments

    @property
    def current(self):
        return self._network

    @property
    def _plugin_context(self):
        return self._original_network

    @property
    def network_segments(self):
        return self._segments


class FakeDriver(object):
    """
    Fake driver which will implement create and delete
    network. Create network will update the global dictionary with
    the address of the device along with vlan and ports to be tagged.
    Example : {'10.10.23.1':{'200':['1/1/1', '1/1/2']}}

    Delete network will delete the corresponding entry from the dictionary.
    """

    def __init__(self, device, error=False):
        self.error = error
        self.device = device
        self.address = device.get('address')

    def create_network(self, vlan_id, ports):
        if self.error:
            raise ml2_exc.MechanismDriverError(method='create_network')
        vlan_map = {}
        vlan_map.update({vlan_id: ports})
        config_map.update({self.address: vlan_map})

    def delete_network(self, vlan_id):
        vlan_map = config_map.pop(self.address, None)
        if vlan_map is None:
            raise ml2_exc.MechanismDriverError(method='delete_network')
        else:
            vlan_map.pop(vlan_id, None)



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import re
import urllib

import netaddr
from neutron.i18n import _LI, _LE
from novaclient import exceptions as nova_exc
from oslo_serialization import jsonutils
from oslo_utils import excutils
import requests

from networking_brocade.vyatta.common import config
from networking_brocade.vyatta.common import exceptions as v_exc
from networking_brocade.vyatta.common import globals as vyatta_globals
from networking_brocade.vyatta.common import parsers
from networking_brocade.vyatta.common import utils as vyatta_utils
from networking_brocade.vyatta.common import vrouter_config


LOG = logging.getLogger(__name__)


class VRouterRestAPIClient(object):
    """Vyatta vRouter REST API Client.

    Uses vRouter REST API to configure vRouter.
    """
    CFG_FMT_GENERIC = 'generic'
    CFG_FMT_COMMANDS = 'commands'

    IF_MAC_ADDRESS = 'mac_address'
    IF_IP_ADDRESS = 'ip_address'
    IF_GATEWAY_IP = 'gateway_ip'

    REST_RETRY_LIMIT = 10
    REST_RETRY_DELAY = 5

    _VROUTER_VSE_MODEL = 54
    _VROUTER_VR_MODEL = 56

    # Floating ip NAT rules will be prioritized before subnet NAT rules.
    # Same rule number is used for both SNAT and DNAT rule.
    _MAX_NAT_FLOATING_IP_RULE_NUM = 4000
    _MAX_NAT_EXCLUDE_RULE_NUM = 8000
    _MAX_NAT_SUBNET_IP_RULE_NUM = 12000

    _EXTERNAL_GATEWAY_DESCR = 'External_Gateway'
    _ROUTER_INTERFACE_DESCR = 'Router_Interface'

    _external_gw_info = None
    _router_if_subnet_dict = {}
    _router_subnet_nat_exclude_dict = {}
    _floating_ip_dict = {}

    # Floating IP NAT rule number counter.
    # It will be incremented in get next method.
    _nat_floating_ip_rule_num = 0

    # SNAT exclude rule number counter.
    # These exclusions need to be prioritized before any
    # subnet NAT rules
    # It will be incremented in get next method
    _nat_exclude_rule_num = _MAX_NAT_FLOATING_IP_RULE_NUM

    # Subnet ip NAT rules are for router interfaces.
    # As we want to prioritize floating ip NAT rules first,
    # subnet rules will start only after floating ip rules.
    # It will be incremented in get next method.
    _nat_subnet_ip_rule_num = _MAX_NAT_EXCLUDE_RULE_NUM

    # Stores the vrouter model
    _vrouter_model = None

    def __init__(self):
        self.address = None

    def connect(self, address):
        """Connects to vRouter using the provided address.

        Retrieves the configuration and updates the cache.
        """
        self.address = address
        LOG.info(_LI("Vyatta vRouter REST API: "
                     "Connecting to vRouter %s"), address)
        self._process_model()
        self._sync_cache()

    def init_router(self, router_name, admin_state_up):
        """
            Configures Router name and Admin State.
        """
        cmd_list = []

        self._set_router_name_cmd(cmd_list, router_name)
        self._set_admin_state_cmd(cmd_list, admin_state_up)

        vyatta_utils.retry(
            self.exec_cmd_batch,
            args=(cmd_list,), exceptions=(v_exc.VRouterOperationError,),
            limit=self.REST_RETRY_LIMIT, delay=self.REST_RETRY_DELAY)

    def update_router(self, router_name=None,
                      admin_state_up=None, external_gateway_info=None):
        """Updates Router name, Admin state, External gateway.

        All the parameters are optional.
        """

        cmd_list = []

        if router_name:
            self._set_router_name_cmd(cmd_list, router_name)

        if admin_state_up is not None:
            self._set_admin_state_cmd(cmd_list, admin_state_up)

        if external_gateway_info is not None:
            given_gw_info = self._get_gw_interface_info(external_gateway_info)
            nat_rules = self._update_gw_config_on_change(given_gw_info,
                                                         cmd_list)
            self._update_gw_cache_info(given_gw_info, nat_rules)
        else:
            self._clear_gw_configuration(cmd_list)
            self._clear_cached_gw_info()

    def add_interface_to_router(self, interface_info):
        """Sets ip address of the ethernet interface.

        Ethernet interface identifier is derived from the given mac-address.
        """

        (if_ip_address,
         eth_if_id) = self._get_ethernet_if_info(interface_info)

        cmd_list = []
        self._set_ethernet_if_cmd(cmd_list,
                                  eth_if_id,
                                  if_ip_address,
                                  self._ROUTER_INTERFACE_DESCR)

        ip_network = netaddr.IPNetwork(if_ip_address)
        router_if_subnet = str(ip_network.cidr)

        # If external gateway was configured before then
        # we need to add SNAT rules
        rule_num = None
        if ip_network.version == 4 and self._external_gw_info is not None:
            rule_num = self._add_snat_rule_for_router_if_cmd(
                cmd_list, router_if_subnet, self._external_gw_info)

        self.exec_cmd_batch(cmd_list)

        # Cache the router interface info using subnet
        if router_if_subnet not in self._router_if_subnet_dict:
            self._router_if_subnet_dict[router_if_subnet] = None

        if self._external_gw_info is not None:
            self._router_if_subnet_dict[router_if_subnet] = rule_num

    def remove_interface_from_router(self, interface_info):
        """Removes ip address of the ethernet interface.

        Ethernet interface identifier is derived from the given mac-address.
        """

        (if_ip_address,
         eth_if_id) = self._get_ethernet_if_info(interface_info)

        cmd_list = []
        self._delete_ethernet_if_cmd(cmd_list,
                                     eth_if_id,
                                     if_ip_address,
                                     self._ROUTER_INTERFACE_DESCR)

        # Check the cache for router interface
        router_if_subnet = self._get_subnet_from_ip_address(if_ip_address)
        if router_if_subnet in self._router_if_subnet_dict:
            # We need to delete the SNAT rule
            nat_rule = self._router_if_subnet_dict[router_if_subnet]
            if nat_rule is not None:
                self._delete_snat_rule_cmd(cmd_list, nat_rule)

        self.exec_cmd_batch(cmd_list)

        # Remove the router interface info from cache
        self._router_if_subnet_dict.pop(router_if_subnet, None)

    def update_interface(self, interface_info):
        if_name = self.get_ethernet_if_id(interface_info.mac_address)
        router_config = self.get_config()
        if_config = router_config.find_interface(if_name)

        old_addrs = set(netaddr.IPNetwork(ip)
                        for ip in if_config.getlist('address'))
        new_addrs = set(interface_info.ip_addresses)

        cmd_list = []

        for ip in old_addrs - new_addrs:
            self._delete_ethernet_ip_cmd(cmd_list, if_name, str(ip))
            # TODO(asaprykin): Configure SNAT

        for ip in new_addrs - old_addrs:
            self._set_ethernet_ip(cmd_list, if_name, str(ip))
            # TODO(asaprykin): Configure SNAT

        self.exec_cmd_batch(cmd_list)

    def assign_floating_ip(self, floating_ip, fixed_ip):
        """Creates SNAT and DNAT rules for given floating ip and fixed ip."""

        if self._external_gw_info is None:
            raise v_exc.VRouterOperationError(
                ip_address=self.address,
                reason='External gateway not configured')

        cmd_list = []

        ext_if_id = self._external_gw_info.get_ethernet_if_id()

        # Get the next NAT rule number and add the NAT rule
        nat_rule_num = self._get_next_nat_floating_ip_rule_num()
        self._add_snat_rule_cmd(cmd_list, nat_rule_num, ext_if_id,
                                fixed_ip, floating_ip)
        self._add_dnat_rule_cmd(cmd_list, nat_rule_num, ext_if_id,
                                floating_ip, fixed_ip)

        # Set the floating ip in external gateway interface
        gw_net = netaddr.IPNetwork(self._external_gw_info.get_ip_address())
        self._set_ethernet_ip(
            cmd_list, self._external_gw_info.get_ethernet_if_id(),
            '{0}/{1}'.format(floating_ip, gw_net.prefixlen))

        self.exec_cmd_batch(cmd_list)

        # Store SNAT and DNAT rule in cache
        dict_key = self._get_floating_ip_key(floating_ip, fixed_ip)
        self._floating_ip_dict[dict_key] = nat_rule_num

    def unassign_floating_ip(self, floating_ip, fixed_ip):
        """Deletes SNAT and DNAT rules for given floating ip and fixed ip."""

        if self._external_gw_info is None:
            raise v_exc.VRouterOperationError(
                ip_address=self.address,
                reason='External gateway not configured')

        cmd_list = []

        # Check the cache for nat rules
        dict_key = self._get_floating_ip_key(floating_ip, fixed_ip)
        if dict_key in self._floating_ip_dict:

            # Get the NAT rules from the cache and delete them
            nat_rule = self._floating_ip_dict[dict_key]
            self._delete_snat_rule_cmd(cmd_list, nat_rule)
            self._delete_dnat_rule_cmd(cmd_list, nat_rule)

            # Delete the floating ip in external gateway interface
            gw_net = netaddr.IPNetwork(self._external_gw_info.get_ip_address())
            self._delete_ethernet_ip_cmd(
                cmd_list, self._external_gw_info.get_ethernet_if_id(),
                '{0}/{1}'.format(floating_ip, gw_net.prefixlen))
        else:
            raise v_exc.VRouterOperationError(
                ip_address=self.address,
                reason='NAT rule not found for floating ip {0}'
                .format(floating_ip))

        self.exec_cmd_batch(cmd_list)

        if dict_key in self._floating_ip_dict:
            self._floating_ip_dict.pop(dict_key)

    def update_static_routes(self, routes_add, routes_del):

        def _get_route_type(dest):
            ip = netaddr.IPNetwork(rule.dest_cidr)
            if ip.version == 4:
                return 'route'
            else:
                return 'route6'

        cmd_list = []
        for rule in routes_add:
            cmd_list.append(SetCmd(
                'protocols/static/{0}/{1}/next-hop/{2}'.format(
                    _get_route_type(rule.dest_cidr),
                    urllib.quote_plus(rule.dest_cidr),
                    urllib.quote_plus(rule.next_hop))))

        for rule in routes_del:
            cmd_list.append(DeleteCmd(
                'protocols/static/{0}/{1}'.format(
                    _get_route_type(rule.dest_cidr),
                    urllib.quote_plus(rule.dest_cidr))))

        self.exec_cmd_batch(cmd_list)

    def disconnect(self):
        self.address = None

    def _rest_call(self, action, uri, custom_headers=None, session=None):
        LOG.debug('Vyatta Router REST Request: {0} {1}'.format(action, uri))
        if session is None:
            session = requests

        auth = tuple(config.VROUTER.vrouter_credentials.split(':'))
        if len(auth) != 2:
            raise v_exc.InvalidParameter(
                cause=_("Invalid vrouter_credentials %s") % len(auth))

        headers = {'Accept': 'application/json',
                   'Content-Length': 0}

        if custom_headers:
            headers.update(custom_headers)

        try:
            uri = 'https://{0}{1}'.format(self.address, uri)
            return session.request(action, uri, auth=auth,
                                   headers=headers, verify=False)
        except requests.ConnectionError:
            LOG.error(_LE('Vyatta vRouter REST API: '
                          'Could not establish HTTP connection to %s'),
                      self.address)
            with excutils.save_and_reraise_exception():
                raise v_exc.VRouterConnectFailure(ip_address=self.address)

    def _get_ethernet_if_info(self, interface_info):
        gw_mac_address = interface_info[self.IF_MAC_ADDRESS]
        gw_ip_address = interface_info[self.IF_IP_ADDRESS]
        gw_if_id = self.get_ethernet_if_id(gw_mac_address)

        return gw_ip_address, gw_if_id

    def _get_gw_interface_info(self, external_gateway_info):
        (gw_ip_address,
         gw_if_id) = self._get_ethernet_if_info(external_gateway_info)
        gw_gateway_ip = external_gateway_info[self.IF_GATEWAY_IP]

        given_gw_info = InterfaceInfo(gw_if_id, gw_ip_address, gw_gateway_ip)
        return given_gw_info

    def _update_gw_config_on_change(self, given_gw_info, cmd_list):
        # Check if the external gw info is already cached.
        # If the given external gw info is not equal to cached gw info
        # then we need to update the existing gw info.
        # So, clear old gw info and set new gw info.
        if (self._external_gw_info is not None and
                given_gw_info != self._external_gw_info):
            LOG.debug("Vyatta vRouter REST API: Cached Gateway info is "
                      "not the same as given gateway info")
            self._delete_external_gateway_if_cmd(
                cmd_list, self._external_gw_info)

        nat_rules = self._set_external_gateway_if_cmd(
            cmd_list, given_gw_info)

        # Execute the configuration commands
        self.exec_cmd_batch(cmd_list)

        return nat_rules

    def _update_gw_cache_info(self, given_gw_info, nat_rules):
        # Cache the external gateway info
        self._external_gw_info = given_gw_info

        # Cache the nat rules
        for router_if_subnet, rule_num in nat_rules.iteritems():
            self._router_if_subnet_dict[router_if_subnet] = rule_num

    def _clear_gw_configuration(self, cmd_list):
        # If external gateway info was cached before
        # then clear the gateway router info
        if self._external_gw_info is not None:
            self._delete_external_gateway_if_cmd(
                cmd_list, self._external_gw_info)
        else:
            raise v_exc.VRouterOperationError(
                ip_address=self.address,
                reason='External gateway not already configured')

        # Execute the configuration commands
        self.exec_cmd_batch(cmd_list)

    def _clear_cached_gw_info(self):
        # Clear the external gateway info from the cache
        self._external_gw_info = None

        # Remove NAT rules for the existing router interfaces
        for router_if_subnet in self._router_if_subnet_dict.keys():
            self._router_if_subnet_dict[router_if_subnet] = None

    def _set_external_gateway_if_cmd(self, cmd_list, gw_info):
        """Sets the external gateway configuration.

        Adds SNAT rules and updates the cache.
        """

        # Set the external gateway ip address
        self._set_ethernet_if_cmd(cmd_list,
                                  gw_info.get_ethernet_if_id(),
                                  gw_info.get_ip_address(),
                                  self._EXTERNAL_GATEWAY_DESCR)

        self._set_system_gateway_cmd(cmd_list, gw_info.get_gateway_ip())

        # Add NAT rules for the existing router interfaces
        nat_rules = {}
        for router_if_subnet in self._router_if_subnet_dict.keys():
            if netaddr.IPNetwork(router_if_subnet).version != 4:
                continue

            rule_num = self._add_snat_rule_for_router_if_cmd(
                cmd_list, router_if_subnet, gw_info)
            nat_rules[router_if_subnet] = rule_num

        return nat_rules

    def _delete_external_gateway_if_cmd(self, cmd_list, gw_info):
        """Sets the external gateway configuration.

        Adds SNAT rules and updates the cache.
        """

        # Remove default gateway
        self._delete_system_gateway_cmd(cmd_list,
                                        gw_info.get_gateway_ip())

        # Delete the external gateway ip address
        self._delete_ethernet_if_cmd(cmd_list,
                                     gw_info.get_ethernet_if_id(),
                                     gw_info.get_ip_address(),
                                     self._EXTERNAL_GATEWAY_DESCR)

        # Remove NAT rules for the existing router interfaces
        for nat_rule in self._router_if_subnet_dict.values():
            self._delete_snat_rule_cmd(cmd_list, nat_rule)

    def _add_snat_rule_for_router_if_cmd(self, cmd_list,
                                         router_if_subnet,
                                         ext_gw_info):

        # Get the next SNAT rule number
        rule_num = self._get_next_nat_subnet_ip_rule_num()

        # Create the SNAT rule and store in the cache
        self._add_snat_rule_cmd(cmd_list,
                                rule_num,
                                ext_gw_info.get_ethernet_if_id(),
                                router_if_subnet,
                                ext_gw_info.get_ip_addr_without_cidr())

        return rule_num

    def _get_snat_exclude_cache_entry(self, src_addr, dest_addr):
        # TODO(sridhar): do we need to support same VPN
        # src-remote pair for multiple
        # external GWs?
        return src_addr + "-" + dest_addr

    def add_snat_exclude_rule(self, cmd_list,
                              ext_if_id, src_addr, dest_addr):

        """Appends vRouter cmds to add a NAT exclude rule
         and caches the rule number in the local 'nat-exclude' dict
         """

        cache_entry = self._get_snat_exclude_cache_entry(src_addr, dest_addr)
        try:
            # if an entry already exist, the nat exclude is already in vRouter
            rule_num = self._router_subnet_nat_exclude_dict[cache_entry]
            LOG.info(_LI('Vyatta vRouter Reuse existing EXCLUDE'
                         ' rule_num : %d'), rule_num)
        except KeyError:
            # Get the next SNAT exclude rule number
            rule_num = self._get_next_nat_exclude_rule_num()
            LOG.info(_LI('Vyatta vRouter Prepare new EXCLUDE'
                         ' rule_num : %d'), rule_num)

        # Create the SNAT rule and store in the cache
        # TODO(sridhar): Unconditionally add for now due client dict bug
        self._add_snat_exclude_rule_cmd(cmd_list,
                                        rule_num,
                                        ext_if_id,
                                        src_addr,
                                        dest_addr)

        LOG.info(_LI('Vyatta vRouter Add Cache EXCLUDE rule_num : %d at %s'),
                 rule_num, cache_entry)
        self._router_subnet_nat_exclude_dict[cache_entry] = rule_num

        return rule_num

    def delete_snat_exclude_rule(self, cmd_list,
                                 ext_if_id, src_addr, dest_addr):
        """Appends vRouter cmds to delete a NAT exclude rule
        and remove the rule number from the local 'nat-exclude' dict
        """

        cache_entry = self._get_snat_exclude_cache_entry(src_addr, dest_addr)
        try:
            nat_rule = self._router_subnet_nat_exclude_dict[cache_entry]
            if nat_rule is not None:
                LOG.info(_LI('Vyatta vRouter: Delete Cache EXCLUDE'
                             ' rule_num : %d at %s'),
                         nat_rule, cache_entry)
                self._delete_snat_rule_cmd(cmd_list, nat_rule)
            else:
                LOG.info(_LI('Vyatta vRouter: Delete Cache EXCLUDE'
                             ' entry at %s doesnt exist'),
                         nat_rule, cache_entry)
        except KeyError:
                LOG.info(_LI('Vyatta vRouter: Delete Cache EXCLUDE:'
                             ' COULD NOT find rule_num at cache entry'
                             ' at %s'), cache_entry)

    def _get_subnet_from_ip_address(self, ip_address):

        ip_network = netaddr.IPNetwork(ip_address)
        # Return subnet with CIDR format
        ip_subnet = str(ip_network.cidr)

        return ip_subnet

    def _get_floating_ip_key(self, floating_ip, fixed_ip):
        """Returns the key to store floating ip and fixed ip combination."""

        return "{0}.{1}".format(floating_ip, fixed_ip)

    def _get_next_nat_floating_ip_rule_num(self):
        """Returns the next NAT rule number for floating ip."""

        if (self._nat_floating_ip_rule_num >=
                self._MAX_NAT_FLOATING_IP_RULE_NUM):
            raise v_exc.VRouterOperationError(
                ip_address=self.address,
                reason='Max NAT Floating IP rule count reached')

        self._nat_floating_ip_rule_num += 1
        return self._nat_floating_ip_rule_num

    def _get_next_nat_exclude_rule_num(self):
        """Returns the next NAT exclude rule number for VPN purposes"""

        if self._nat_exclude_rule_num >= self._MAX_NAT_EXCLUDE_RULE_NUM:
            raise v_exc.VRouterOperationError(
                ip_address=self.address,
                reason='Max NAT Exclude rule count reached')

        self._nat_exclude_rule_num += 1
        return self._nat_exclude_rule_num

    def _get_next_nat_subnet_ip_rule_num(self):
        """Returns the next NAT rule number for subnet ip."""

        if self._nat_subnet_ip_rule_num >= self._MAX_NAT_SUBNET_IP_RULE_NUM:
            raise v_exc.VRouterOperationError(
                ip_address=self.address,
                reason='Max NAT Subnet IP rule count reached')

        self._nat_subnet_ip_rule_num += 1
        return self._nat_subnet_ip_rule_num

    def _get_admin_state(self):
        """Retrieves Admin State."""
        output = self._show_cmd("ip/forwarding")
        LOG.info(_LI('Vyatta vRouter status : %s'), output)
        return "IP forwarding is on" in output

    def _get_nat_cmd(self):

        return 'service/nat' if (self._vrouter_model ==
                                 self._VROUTER_VR_MODEL) else 'nat'

    def _add_snat_rule_cmd(self, cmd_list, rule_num, ext_if_id,
                           src_addr, translation_addr):
        """Creates SNAT rule with the given parameters."""

        nat_cmd = self._get_nat_cmd()

        # Execute the commands
        cmd_list.append(
            SetCmd("{0}/source/rule/{1}".format(nat_cmd, rule_num)))
        cmd_list.append(SetCmd("{0}/source/rule/{1}/outbound-interface/{2}"
                               .format(nat_cmd, rule_num, ext_if_id)))
        cmd_list.append(SetCmd("{0}/source/rule/{1}/source/address/{2}"
                               .format(nat_cmd, rule_num,
                                       urllib.quote_plus(src_addr))))
        cmd_list.append(SetCmd("{0}/source/rule/{1}/translation/address/{2}"
                               .format(nat_cmd, rule_num,
                                       urllib.quote_plus(translation_addr))))

    def _add_snat_exclude_rule_cmd(self, cmd_list, rule_num, ext_if_id,
                                   src_addr, dest_addr):
        """Create SNAT exclude rule between tenant networks and
        remote VPN CIDR
        """

        nat_cmd = self._get_nat_cmd()

        LOG.info(_LI('Vyatta vRouter Adding EXCLUDE rule_num'
                     ' : %s, ext_if_id %s, src_addr %s, dest_addr %s'),
                 rule_num, ext_if_id, src_addr, dest_addr)

        # Execute the commands
        cmd_list.append(SetCmd("{0}/source/rule/{1}"
                               .format(nat_cmd, rule_num)))
        cmd_list.append(SetCmd("{0}/source/rule/{1}/exclude"
                               .format(nat_cmd, rule_num)))
        cmd_list.append(SetCmd("{0}/source/rule/{1}/outbound-interface/{2}"
                               .format(nat_cmd, rule_num, ext_if_id)))
        cmd_list.append(SetCmd("{0}/source/rule/{1}/source/address/{2}"
                               .format(nat_cmd, rule_num,
                                       urllib.quote_plus(src_addr))))
        cmd_list.append(SetCmd("{0}/source/rule/{1}/destination/address/{2}"
                               .format(nat_cmd, rule_num,
                                       urllib.quote_plus(dest_addr))))

    def _add_dnat_rule_cmd(self, cmd_list, rule_num, ext_if_id,
                           dest_addr, translation_addr):
        """Creates DNAT rule with the given parameters."""

        nat_cmd = self._get_nat_cmd()

        # Execute the commands
        cmd_list.append(
            SetCmd("{0}/destination/rule/{1}".format(nat_cmd, rule_num)))
        cmd_list.append(SetCmd("{0}/destination/rule/{1}/inbound-interface/{2}"
                               .format(nat_cmd, rule_num, ext_if_id)))
        cmd_list.append(SetCmd("{0}/destination/rule/{1}/destination/"
                               "address/{2}".format(
                                   nat_cmd, rule_num,
                                   urllib.quote_plus(dest_addr))))
        cmd_list.append(SetCmd("{0}/destination/rule/{1}/translation/"
                               "address/{2}".format(
                                   nat_cmd, rule_num,
                                   urllib.quote_plus(translation_addr))))

    def _delete_snat_rule_cmd(self, cmd_list, rule_num):
        """Deletes the given SNAT rule."""

        cmd_list.append(DeleteCmd("{0}/source/rule/{1}".
                                  format(self._get_nat_cmd(), rule_num)))

    def _delete_dnat_rule_cmd(self, cmd_list, rule_num):
        """Deletes the given DNAT rule."""

        cmd_list.append(DeleteCmd("{0}/destination/rule/{1}".
                                  format(self._get_nat_cmd(), rule_num)))

    def _set_admin_state_cmd(self, cmd_list, admin_state):
        """Sets Admin State using command."""

        if admin_state:
            if not self._get_admin_state():
                cmd_list.append(DeleteCmd("system/ip/disable-forwarding"))
        else:
            if self._get_admin_state():
                cmd_list.append(SetCmd("system/ip/disable-forwarding"))

    def get_vrouter_configuration(self, mode=CFG_FMT_GENERIC):
        cmd = ['configuration']
        if mode == self.CFG_FMT_GENERIC:
            pass
        elif mode == self.CFG_FMT_COMMANDS:
            cmd.append('commands')
        else:
            raise v_exc.InvalidParameter(
                cause='unsupported configuration dump format')

        cmd = '/'.join(cmd)
        return self._show_cmd(cmd)

    def get_vpn_ipsec_sa(self, peer=None, tunnel=None):
        assert not tunnel or peer

        cmd = ['vpn', 'ipsec', 'sa']
        if peer:
            cmd.append('peer')
            cmd.append(urllib.quote_plus(peer))
        if tunnel:
            cmd.append('tunnel')
            cmd.append(urllib.quote_plus(tunnel))
        cmd = '/'.join(cmd)
        return self._show_cmd(cmd)

    def get_ethernet_if_id(self, mac_address):
        """Uses show command output to find the ethernet interface."""

        LOG.debug('Vyatta vRouter:get_ethernet_if_id. Given MAC {0}'
                  .format(repr(mac_address)))
        iface = self._find_interface(mac_address)
        return iface['name']

    def get_config(self):
        raw_config = self._show_cmd('configuration/all')
        return vrouter_config.RouterConfig.from_string(raw_config)

    def _get_interface_cmd(self):
        if self._vrouter_model == self._VROUTER_VR_MODEL:
            return "dataplane"
        else:
            return "ethernet"

    def _set_ethernet_ip(self, cmd_list, if_id, ip_address):
        """Sets ip address to an ethernet interface."""

        if_cmd = self._get_interface_cmd()

        cmd_list.append(SetCmd("interfaces/{0}/{1}/address/{2}"
                               .format(if_cmd, if_id,
                                       urllib.quote_plus(ip_address))))

    def _set_ethernet_if_cmd(self, cmd_list, if_id,
                             ip_address, descr):
        """Sets ip address and description of an ethernet interface."""

        if_cmd = self._get_interface_cmd()

        # Execute the commands
        cmd_list.append(SetCmd("interfaces/{0}/{1}/address/{2}"
                               .format(if_cmd, if_id,
                                       urllib.quote_plus(ip_address))))
        cmd_list.append(SetCmd("interfaces/{0}/{1}/description/{2}"
                               .format(if_cmd, if_id,
                                       urllib.quote_plus(descr))))

    def _delete_ethernet_ip_cmd(self, cmd_list, if_id, ip_address):
        """Deletes ip address from an ethernet interface."""

        if_cmd = self._get_interface_cmd()

        cmd_list.append(DeleteCmd("interfaces/{0}/{1}/address/{2}"
                                  .format(if_cmd, if_id,
                                          urllib.quote_plus(ip_address))))

    def _delete_ethernet_if_cmd(self, cmd_list, if_id,
                                ip_address, descr):
        """Deletes ip address and description of an ethernet interface."""

        if_cmd = self._get_interface_cmd()

        # Execute the commands
        cmd_list.append(DeleteCmd("interfaces/{0}/{1}/address/{2}"
                                  .format(if_cmd, if_id,
                                          urllib.quote_plus(ip_address))))
        cmd_list.append(DeleteCmd("interfaces/{0}/{1}/description/{2}"
                                  .format(if_cmd, if_id,
                                          urllib.quote_plus(descr))))
        cmd_list.append(DeleteCmd("interfaces/{0}/{1}".
                                  format(if_cmd, if_id)))

    def _set_router_name_cmd(self, cmd_list, router_name):
        """Configures router name using command."""
        if '_' in router_name:
            router_name = router_name.replace('_', '-')

        cmd_list.append(SetCmd("system/host-name/{0}".
                               format(urllib.quote_plus(router_name))))

    def _set_system_gateway_cmd(self, cmd_list, gateway_ip):

        cmd_list.append(SetCmd("protocols/static/route/{0}/next-hop/{1}".
                               format(urllib.quote_plus('0.0.0.0/0'),
                                      urllib.quote_plus(gateway_ip))))

    def _delete_system_gateway_cmd(self, cmd_list, gateway_ip):

        cmd_list.append(DeleteCmd("protocols/static/route/{0}".
                                  format(urllib.quote_plus('0.0.0.0/0'))))

    def _configure_cmd(self, cmd_type, cmd):
        """Executes the given configuration command.

        Commits and Saves the configuration changes to the startup config.
        """

        self.configure_cmd_list(cmd_type, [cmd])

    def exec_cmd_batch(self, user_cmd_list):
        """Executes the given configuration command list.

        Commits and Saves the configuration changes to the startup config.
        """
        response = self._rest_call("POST", "/rest/conf")
        self._check_response(response)

        config_url = response.headers['Location']
        if config_url is None:
            raise v_exc.VRouterOperationError(
                ip_address=self.address,
                reason='REST API configuration URL is null')

        config_url = "/" + config_url
        for user_cmd in user_cmd_list:
            url = user_cmd.make_url(config_url)
            LOG.debug(
                "Vyatta vRouter REST API: Config command %s", url)
            response = self._rest_call("PUT", url)
            self._check_response(response, config_url)

        response = self._rest_call(
            "POST", config_url + "/commit")
        LOG.debug("Vyatta vRouter REST API: %s/commit", config_url)
        self._check_response(response, config_url)

        response = self._rest_call(
            "POST", config_url + "/save")
        LOG.debug("Vyatta vRouter REST API: %s/save", config_url)
        self._check_response(response, config_url)

        response = self._rest_call("DELETE", config_url)
        self._check_response(response)

    def _check_response(self, response, config_url=None, session=None):

        if session is None:
            session = requests

        if response.status_code not in (requests.codes.OK,
                                        requests.codes.CREATED):
            LOG.error(_LE('Vyatta vRouter REST API: Response Status : '
                      '%(status)s Reason: %(reason)s') %
                      {'status': response.status_code,
                       'reason': response.reason})

            if config_url is not None:
                self._rest_call("DELETE", config_url, session=session)

            raise v_exc.VRouterOperationError(
                ip_address=self.address, reason=response.reason)

    def _get_config_cmd(self, user_cmd):
        """Executes the given "get config" command."""

        response = self._rest_call("POST", "/rest/conf")
        self._check_response(response)

        config_url = response.headers['Location']
        if config_url is None:
            raise v_exc.VRouterOperationError(
                ip_address=self.address,
                reason='REST API Configuration URL is None')
        config_url = "/" + config_url
        config_cmd = '{0}/{1}/'.format(config_url, user_cmd)
        response = self._rest_call("GET", config_cmd)
        self._check_response(response)
        data = jsonutils.loads(response.text)
        self._rest_call("DELETE", config_url)
        return data

    def _show_cmd(self, user_cmd):
        # TODO(asaprykin): Need to verify error handling

        op_cmd = '/rest/op/show/{0}'.format(user_cmd)
        response = self._rest_call("POST", op_cmd)
        self._check_response(response)

        op_url = response.headers['Location']
        if op_url is None:
            raise v_exc.VRouterOperationError(
                ip_address=self.address, reason='REST API Op URL is None')

        op_url = "/" + op_url

        output = []

        while True:
            response = self._rest_call('GET', op_url)
            if response.status_code == requests.codes.GONE:
                break

            if response.text:
                output.append(response.text)

        self._rest_call("DELETE", op_url)
        return ''.join(output)

    def _process_model(self):

        model = None
        show_output = self._show_cmd("version")
        LOG.debug('Vyatta vRouter REST API: Version output : %s',
                  show_output)
        if show_output is not None:
            ma = re.compile(".+Description.+Brocade Vyatta\D+(\d+).+",
                            re.DOTALL)
            result = ma.match(show_output)
            LOG.debug('Vyatta vRouter REST API: Result : %s', result)
            if result is not None:
                model_str = result.group(1)
                LOG.debug('Vyatta vRouter REST API: Result : %s',
                          model_str)
                model = int(model_str) / 100
                LOG.debug('Vyatta vRouter REST API: Result : %s',
                          model)
                if model in (self._VROUTER_VSE_MODEL, self._VROUTER_VR_MODEL):
                    self._vrouter_model = model

        LOG.debug('Vyatta vRouter REST API: Version : %s',
                  self._vrouter_model)
        if self._vrouter_model is None:
            raise v_exc.VRouterOperationError(
                ip_address=self.address,
                reason='Unable to process vRouter model info: {0}'
                .format(model))

    def _sync_cache(self):

        show_output = self._show_cmd("configuration/all")

        system_gw = None
        gateway_str = self._get_config_block("protocols", show_output)
        if gateway_str is not None:
            system_gw = self._parse_system_gateway(gateway_str)

        interfaces_str = self._get_config_block("interfaces", show_output)
        if interfaces_str is not None:
            self._process_interfaces(interfaces_str, system_gw)

        if self._vrouter_model == self._VROUTER_VR_MODEL:
            show_output = self._get_config_block("service", show_output)

        nat_str = self._get_config_block("nat", show_output)
        if nat_str is not None:
            self._process_source_nat_rules(nat_str)

        LOG.info(_LI("Vyatta vRouter cache ext gw %s"),
                 self._external_gw_info)
        LOG.info(_LI("Vyatta vRouter cache router if dict %s"),
                 self._router_if_subnet_dict)
        LOG.info(_LI("Vyatta vRouter cache floating ip dict %s"),
                 self._floating_ip_dict)
        LOG.info(_LI("Vyatta vRouter cache router nat-exclude dict %s"),
                 self._router_subnet_nat_exclude_dict)
        LOG.info(_LI("Vyatta vRouter cache NAT floating ip %s"),
                 self._nat_floating_ip_rule_num)
        LOG.info(_LI("Vyatta vRouter cache NAT subnet ip %s"),
                 self._nat_subnet_ip_rule_num)
        LOG.info(_LI("Vyatta vRouter cache NAT exclude rule num %s"),
                 self._nat_exclude_rule_num)

    def _parse_system_gateway(self, search_str):

        system_gw_ip = None
        ma = re.compile(".+static.+route.+next-hop ([^ \n]+).+", re.DOTALL)
        result = ma.match(search_str)
        if result is not None:
            system_gw_ip = result.group(1)
        return system_gw_ip

    def _process_interfaces(self, search_str, system_gw_ip):

        for paragraph in search_str.split('}'):
            ma = re.compile(
                ".+ethernet (eth\d+).+address ([^ \n]+).+description ([^ \n]+)"
                ".+", re.DOTALL)
            result = ma.match(paragraph)
            if result is not None:
                eth_if_id = result.group(1)
                ip_address = result.group(2)
                description = result.group(3)
                if description == self._EXTERNAL_GATEWAY_DESCR:
                    ext_gw_info = InterfaceInfo(eth_if_id,
                                                ip_address, system_gw_ip)
                    self._external_gw_info = ext_gw_info
                elif description == self._ROUTER_INTERFACE_DESCR:
                    # Cache the router interface info using subnet
                    router_if_subnet = self._get_subnet_from_ip_address(
                        ip_address)
                    self._router_if_subnet_dict[router_if_subnet] = None

    def _process_source_nat_rules(self, search_str):

        for paragraph in search_str.split('rule'):
            ma = re.compile(
                ".(\d+).+outbound-interface.+source.+address ([^ \n]+)"
                ".+translation.+address ([^ \n]+).+", re.DOTALL)
            result = ma.match(paragraph)
            if result is not None:
                rule_num = int(result.group(1))
                src_addr = result.group(2)
                translation_addr = result.group(3)
                if (self._MAX_NAT_EXCLUDE_RULE_NUM < rule_num <
                   self._MAX_NAT_SUBNET_IP_RULE_NUM and
                   src_addr in self._router_if_subnet_dict):
                    # Cache the SNAT rule for router interface
                    self._router_if_subnet_dict[src_addr] = rule_num
                    self._nat_subnet_ip_rule_num = rule_num
                elif (self._MAX_NAT_FLOATING_IP_RULE_NUM < rule_num <
                    self._MAX_NAT_EXCLUDE_RULE_NUM and
                    src_addr in self._router_if_subnet_dict):
                    # TODO(sridhar): Add parsing code for exclude nat rule
                    pass
                elif rule_num < self._MAX_NAT_FLOATING_IP_RULE_NUM:
                    self._nat_floating_ip_rule_num = rule_num
                    floating_ip = translation_addr
                    fixed_ip = src_addr
                    # Store SNAT and DNAT rule in cache
                    dict_key = self._get_floating_ip_key(floating_ip, fixed_ip)
                    self._floating_ip_dict[dict_key] = rule_num

    def _get_config_block(self, input_str, search_str):

        if search_str is not None:
            index = search_str.find(input_str)
            if index >= 0:
                block_start = search_str[index + len(input_str):]
                block_str = []
                for line in block_start.split('\n'):
                    if line.startswith('}'):
                        break
                    block_str.append(line)
                return ''.join(block_str)

        return None

    def _get_interfaces(self):
        output = self._show_cmd('interfaces/detail')
        return parsers.parse_interfaces(output)

    def _find_interface(self, mac_address):
        mac_address = mac_address.strip().lower()
        ifaces = self._get_interfaces()
        for iface in ifaces:
            if iface['mac_address'] == mac_address:
                return iface

        raise v_exc.VRouterOperationError(
            ip_address=self.address,
            reason='Ethernet interface with Mac-address {0} does not exist'
            .format(mac_address))


class ClientsPool(object):
    _client_factory = VRouterRestAPIClient

    def __init__(self, compute_client):
        # TODO(dbogun): avoid dependency from nova client
        self._compute_client = compute_client
        self._active_connections = dict()

    def get_by_address(self, router_id, address):
        try:
            client = self._active_connections[router_id]
        except KeyError:
            self._active_connections[router_id] = client = \
                self._client_factory()
            client.connect(address)
        return client

    def get_by_db_lookup(self, router_id, context):
        try:
            client = self._active_connections[router_id]
        except KeyError:
            self._active_connections[router_id] = client = \
                self._make_connection(context, router_id)
        return client

    def _make_connection(self, context, router_id):
        LOG.debug('Vyatta vRouter Driver::Get router driver')

        try:
            router = self._compute_client.servers.get(router_id)
        except nova_exc.ClientException as ex:
            LOG.error(_LE(
                'Unable to find Vyatta vRouter instance {0}. '
                'Exception {1}').format(router_id, ex))
            raise v_exc.InvalidVRouterInstance(router_id=router_id)

        network = vyatta_globals.get_management_network(context)
        network = context.session.merge(network)
        if network in context.session.new:
            raise v_exc.InvalidInstanceConfiguration(
                cause='Unable to find management network')

        LOG.debug('Vyatta vRouter Management network: {0}'.format(
            network['name']))
        try:
            address_map = router.addresses[network['name']]
            address = address_map[0]['addr']
        except (KeyError, IndexError):
            raise v_exc.InvalidVRouterInstance(router_id=router_id)

        return self._client_by_address(address)

    def _client_by_address(self, address):
        # Initialize vRouter API
        try:
            client = self._client_factory()
            client.connect(address)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE('Vyatta vRouter Driver: vRouter {0} '
                              'Connection exception {1}').format(
                    address, ex))
            raise
        return client


# REST API commands
class UserCmd(object):

    def __init__(self, cmd_type, cmd):
        self.cmd_type = cmd_type
        self.cmd = cmd

    def __repr__(self):
        return '{0} {1!r}'.format(self.cmd_type, self.cmd)

    def __eq__(self, other):
        if not isinstance(other, UserCmd):
            return NotImplemented
        return (self.cmd_type, self.cmd) == (other.cmd_type, other.cmd)

    def __ne__(self, other):
        return not self.__eq__(other)

    def make_url(self, prefix):
        url = (prefix, self.cmd_type, self.cmd)
        return '/'.join(url)


class SetCmd(UserCmd):

    def __init__(self, cmd):
        super(SetCmd, self).__init__("set", cmd)


class DeleteCmd(UserCmd):

    def __init__(self, cmd):
        super(DeleteCmd, self).__init__("delete", cmd)


class InterfaceInfo(object):

    """Class for storing interface related info."""
    def __init__(self, ethernet_if_id, ip_address,
                 gateway_ip=None):
        self._ethernet_if_id = ethernet_if_id
        self._ip_address = ip_address
        self._gateway_ip = gateway_ip
        self._ip_addr_without_cidr = None

    def get_ethernet_if_id(self):
        return self._ethernet_if_id

    def get_ip_address(self):
        return self._ip_address

    def get_ip_addr_without_cidr(self):
        if self._ip_addr_without_cidr is None:
            # Find the subnet
            ip_network = netaddr.IPNetwork(self._ip_address)
            # Without CIDR format
            self._ip_addr_without_cidr = str(ip_network.ip)

        return self._ip_addr_without_cidr

    def get_gateway_ip(self):
        return self._gateway_ip

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return 'Eth if:{0} IP:{1} GW:{2}'.format(self._ethernet_if_id,
                                                 self._ip_address,
                                                 self._gateway_ip)

    def __repr(self):
        return self.__str__()



# Copyright 2015 Brocade Communications System, Inc..
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


cfg.CONF.register_opts([
    cfg.StrOpt('tenant_admin_name', help=_('Name of tenant admin user.')),
    cfg.StrOpt('tenant_admin_password', secret=True,
               help=_('Tenant admin password.')),
    cfg.StrOpt('tenant_id',
               help=_('UUID of tenant that holds Vyatta vRouter instances.')),
    cfg.StrOpt('image_id',
               help=_('Nova image id for instances of Vyatta vRouter.')),
    cfg.StrOpt('flavor', default=2,
               help=_('Nova VM flavor for instances of Vyatta vRouter.')),
    cfg.StrOpt('management_network_id',
               help=_('Vyatta vRouter management network id.')),
    cfg.StrOpt('vrouter_credentials', default="vyatta:vyatta",
               help=_('Vyatta vRouter login credentials')),
    cfg.IntOpt('nova_poll_interval', default=5,
               help=_('Number of seconds between consecutive Nova queries '
                      'when waiting for router instance status change.')),
    cfg.IntOpt('nova_spawn_timeout', default=300,
               help=_('Number of seconds to wait for Nova to activate '
                      'instance before setting resource to error state.')),
    cfg.IntOpt('vrouter_poll_interval', default=5,
               help=_('Number of seconds between consecutive Vyatta vRouter '
                      'queries when waiting for router instance boot.')),
    cfg.IntOpt('vrouter_boot_timeout', default=300,
               help=_('Number of seconds to wait for Vyatta vRouter to boot '
                      'before setting resource to error state.')),
    cfg.StrOpt('keystone_url', help=_('Keystone URL.')),
], "VROUTER")


# setup shortcuts
CONF = cfg.CONF
VROUTER = cfg.CONF.VROUTER



# Copyright 2015 Brocade Communications Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Parses the brocade ethernet configuration template
"""

from oslo_config import cfg


class ML2BrocadeConfig(object):

    """
    Parse the configuration template for Brocade Ethernet ML2 plugin
    """

    def __init__(self):
        self._brocade_switches = None
        self._brocade_dict = {}
        self._physnet_dict = {}

    def create_brocade_dictionary(self, isL2=True):
        """
        Create the brocade dictionary.
        Read data from the ml2_conf_brocade.ini.

        :returns: Two dictionaries, one contains device details with device
            name as key and the details as value. Second dictionary contains
            physical_network as key and list of device names as value
        :param:isL2: Boolean value based on the router type L2 or L3
        """
        if isL2:
            self._brocade_switches = (cfg.CONF.ml2_brocade_fi_ni.
                                      switch_names)
        else:
            self._brocade_switches = (cfg.CONF.l3_brocade_mlx.
                                      switch_names)
        switches = self._brocade_switches.split(',')
        for switch in switches:
            switch_info = {}
            switch = switch.strip()
            for key, value in cfg.CONF._get(switch).items():
                value = value.strip()
                switch_info.update({key: value})
                if "physical_networks" in key:
                    switch_names = self._physnet_dict.get(value)
                    if switch_names is None:
                        switch_names = []
                        switch_names.append(switch)
                    else:
                        switch_names.append(switch)

                    self._physnet_dict.update({value: switch_names})
            if not isL2:
                switch_info.update({'transport': 'SSH'})
                switch_info.update({'ostype': 'NI'})
            self._brocade_dict.update({switch: switch_info})
        return self._brocade_dict, self._physnet_dict



# Copyright 2015 Brocade Communications System, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
    Connector class used to connect to device.
    Decides which connector to use - TELNET or SSH, based on
    the argument passed
"""

from oslo_utils import importutils


class ConnectorFactory(object):
    """
    Connector class used to connect to device.
    Decides which connector to use - TELNET or SSH, based on
    the argument passed
    """

    connectors = {"SSH": "networking_brocade.mlx.ml2.ssh_connector."
                  "SSHConnector",
                  "TELNET": "networking_brocade.mlx.ml2."
                            "telnet_connector.TelnetConnector"
                  }

    def get_connector(self, device_info):
        """
        Returns the connector based on the transport to be used

        :param:device_info: Device details in the form of dictionary
        :returns: Connector for the device based on the transport parameter.
            Currently only SSH is supported and it is the default connector
            that is returned.
        """
        transport = device_info.get('transport')
        connector = self.connectors.get(transport)
        connection = importutils.import_object(connector, device_info)
        return connection



# Copyright 2015 Brocade Communications Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Returns the driver based on the firmware version of the device
"""
from neutron.i18n import _LE
from neutron.i18n import _LI
from oslo_log import log as logging
from oslo_utils import importutils

LOG = logging.getLogger(__name__)
CONNECTION_FACTORY = ("networking_brocade.mlx.ml2.connector_factory."
                      "ConnectorFactory")
FI_DRIVER = "networking_brocade.mlx.ml2.fi_ni.fi_driver.FastIronDriver"
NI_DRIVER = "networking_brocade.mlx.ml2.fi_ni.ni_driver.NetIronDriver"
NETIRON = "NetIron"
FASTIRON = "ICX"
FI = "FI"
NI = "NI"


class BrocadeDriverFactory(object):

    """
    Factory class that decides which driver to use based on the
    device type. It uses FastIron driver for ICX devices and
    NetIron driver for MLX devices
    """

    def get_driver(self, device):
        """
        Returns the driver based on the firmware.

        :param:device: A dictionary which has the device details
        :returns: Appropriate driver for the device based on the firmware
            version, None otherwise
        :raises: Exception
        """
        driver = None
        address = device.get('address')
        os_type = device.get('ostype')
        if os_type == FI:
            driver = importutils.import_object(FI_DRIVER, device)
        elif os_type == NI:
            driver = importutils.import_object(NI_DRIVER, device)
        else:
            connector = importutils.import_object(CONNECTION_FACTORY
                                                  ).get_connector(device)
            connector.connect()
            version = connector.get_version()
            connector.close_session()
            if NETIRON in version:
                LOG.info(
                    _LI("OS Type of the device %(host)s is as NetIron"),
                    {'host': address})
                driver = importutils.import_object(NI_DRIVER, device)
                device.update({'ostype': NI})
            elif FASTIRON in version:
                LOG.info(
                    _LI("OS Type of the device %(host)s is as FastIron"),
                    {'host': device.get('address')})
                driver = importutils.import_object(FI_DRIVER, device)
                device.update({'ostype': FI})
            else:
                LOG.exception(_LE("Brocade Driver Factory: failed to "
                                  "identify device type for device="
                                  "%(device)s"), {'device': address})

                raise Exception("Unsupported firmware %(firmware)s for device "
                                "%(host)s", {'firmware': version,
                                             'host': address})

        return driver



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from eventlet import greenthread
import netaddr
import six


RouteRule = collections.namedtuple('RouteRule', 'dest_cidr, next_hop')


def retry(fn, args=None, kwargs=None, exceptions=None, limit=1, delay=0):
    args = args or []
    kwargs = kwargs or {}

    while limit > 0:
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            if not exceptions or not isinstance(e, exceptions):
                raise
        if delay:
            greenthread.sleep(delay)
        limit -= 1
    raise


class MultiDict(collections.MutableMapping):

    def __init__(self, mapping=None):
        self._items = {}

        if isinstance(mapping, MultiDict):
            for key, value in mapping.lists():
                self._items[key] = value[:]
        elif isinstance(mapping, dict):
            for key, value in six.iteritems(mapping):
                self._items[key] = [value]
        elif mapping is not None:
            for key, value in mapping:
                self._items.setdefault(key, []).append(value)

    def __getitem__(self, key):
        return self._items[key][0]

    def __setitem__(self, key, value):
        self._items[key] = [value]

    def __delitem__(self, key):
        del self._items[key]

    def __len__(self):
        return len(self._items)

    def __iter__(self):
        return six.iterkeys(self._items)

    def __repr__(self):
        items = []
        for key, lst in six.iteritems(self._items):
            for item in lst:
                items.append((key, item))
        return 'MultiDict {0!r}'.format(items)

    def add(self, key, value):
        self._items.setdefault(key, []).append(value)

    def getlist(self, key, default=None):
        try:
            return self._items[key]
        except KeyError:
            return default or []

    def setlist(self, key, value):
        self._items[key] = list(value)

    def setlistdefault(self, key, default_list=None):
        if key in self._items:
            default_list = self._items[key]
        else:
            if default_list is None:
                default_list = []
            else:
                default_list = list(default_list)
            self._items[key] = default_list

        return default_list

    def copy(self):
        return self.__class__(self)

    def lists(self):
        return six.iteritems(self._items)

    def listvalues(self):
        return six.itervalues(self._items)


class InterfaceInfo(object):

    def __init__(self, name=None, ip_addresses=None,
                 mac_address=None, gateway=None):

        self.name = name
        self.mac_address = mac_address
        self.gateway = gateway

        self.ip_addresses = []
        for addr in ip_addresses:
            if isinstance(addr, str):
                addr = netaddr.IPNetwork(addr)
            elif not isinstance(addr, netaddr.IPNetwork):
                raise TypeError('IP address {0!r} should be '
                                'str or netaddr.IPNetwork')
            self.ip_addresses.append(addr)



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient.v1_1 import client as novaclient

from neutron.agent.l3 import agent as l3_agent
from neutron.agent.l3 import router_info
from neutron.agent import l3_agent as entry
from neutron.common import constants as l3_constants

from networking_brocade.vyatta.common import config as vyatta_config
from networking_brocade.vyatta.common import exceptions as v_exc
from networking_brocade.vyatta.vrouter import client as vyatta_client


_KEY_VYATTA_EXTRA_DATA = '_vyatta'
_KEY_MANAGEMENT_IP_ADDRESS = 'management_ip_address'


class RouterInfo(router_info.RouterInfo):

    def add_floating_ip(self, fip, interface_name, device):
        pass

    def initialize(self, process_monitor):
        pass

    def process(self, agent):
        pass

    def delete(self, agent):
        self.router['gw_port'] = None
        self.router[l3_constants.INTERFACE_KEY] = []
        self.router[l3_constants.FLOATINGIP_KEY] = []
        self.process(agent)


class L3AgentMiddleware(l3_agent.L3NATAgentWithStateReport):
    def __init__(self, host, conf=None):
        super(L3AgentMiddleware, self).__init__(host, conf)

        compute_client = novaclient.Client(
            vyatta_config.VROUTER.tenant_admin_name,
            vyatta_config.VROUTER.tenant_admin_password,
            auth_url=vyatta_config.CONF.nova_admin_auth_url,
            service_type="compute",
            tenant_id=vyatta_config.VROUTER.tenant_id)
        self._vyatta_clients_pool = vyatta_client.ClientsPool(compute_client)

    def get_router(self, router_id):
        try:
            router = self.router_info[router_id]
        except KeyError:
            raise v_exc.InvalidL3AgentStateError(description=_(
                'L3 agent have no info about reouter id={0}').format(
                    router_id))
        return router.router

    def get_router_client(self, router_id):
        router = self.get_router(router_id)
        try:
            address = router[_KEY_VYATTA_EXTRA_DATA]
            address = address[_KEY_MANAGEMENT_IP_ADDRESS]
        except KeyError:
            raise v_exc.CorruptedSystemError(
                description=('router {0} does not contain vyatta vrouter '
                             'management ip address').format(router_id))
        return self._vyatta_clients_pool.get_by_address(router_id, address)

    def _create_router(self, router_id, router):
        if router.get('distributed') or router.get('ha'):
            raise v_exc.DvrOrHaRouterNotSupported(router_id=router_id)

        kwargs = {
            'router_id': router_id,
            'router': router,
            'use_ipv6': self.use_ipv6,
            'agent_conf': self.conf,
            'interface_driver': self.driver,
        }
        return RouterInfo(**kwargs)

    def _get_router_info_list_for_tenant(self, router_ids, tenant_id):
        """Returns the list of router info objects on which to apply the fw."""
        router_info_list = []
        # Pick up namespaces for Tenant Routers
        for rid in router_ids:
            # for routers without an interface - get_routers returns
            # the router - but this is not yet populated in router_info
            if rid not in self.router_info:
                continue
            router_info_list.append(self.router_info[rid])
        return router_info_list


def main():
    entry.main(
        manager='networking_brocade.vyatta.common.l3_agent.L3AgentMiddleware')



# Copyright 2015 Brocade Communications System, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Implementation of Telnet Connector """

import telnetlib

from networking_brocade.mlx.ml2.device_connector import (
    DeviceConnector as DevConn)
from neutron.i18n import _LE
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

TELNET_PORT = 23
LOGIN_USER_TOKEN = "Name:"
LOGIN_PASS_TOKEN = "Password:"
PATTERN_EN_AUTH = '.*\\r\\n.*Name\\:$'
SUPER_USER_AUTH = '^Password\\:$'
TERMINAL_LENGTH = "terminal length 0"

END_OF_LINE = "\r"
TELNET_TERMINAL = ">"
CONFIGURE_TERMINAL = "#"

MIN_TIMEOUT = 2
AVG_TIMEOUT = 4
MAX_TIMEOUT = 8


class TelnetConnector(DevConn):

    """
    Uses Telnet to connect to device
    """

    def connect(self):
        """
        Connect to device via Telnet
        """
        try:
            self.connector = telnetlib.Telnet(host=self.host, port=TELNET_PORT)
            self.connector.read_until(LOGIN_USER_TOKEN, MIN_TIMEOUT)
            self.connector.write(self.username + END_OF_LINE)
            self.connector.read_until(LOGIN_PASS_TOKEN, AVG_TIMEOUT)
            self.connector.write(self.password + END_OF_LINE)
            self.connector.read_until(TELNET_TERMINAL, MAX_TIMEOUT)
        except Exception as e:
            LOG.exception(_LE("Connect failed to switch %(host)s with error"
                              " %(error)s"),
                          {'host': self.host, 'error': e.args})
            raise Exception(_("Connection Failed"))

    def write(self, command):
        """
        Write from input stream to device

        :param:command: Command to be executed on the device
        """
        self.connector.write(command)

    def read_response(self, read_lines=True):
        """Read the response from the output stream.

        :param:read_lines: This is used only by the SSH connector.
        :returns: Response from the device as list.
        """
        return self.connector.read_until(CONFIGURE_TERMINAL, MIN_TIMEOUT)

    def close_session(self):
        """Close TELNET session."""
        if self.connector:
            self.connector.close()
            self.connector = None

    def send_exit(self, count):
        """No operation. Used by SSH connector only."""
        pass



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import time

from eventlet import greenthread
from oslo_log import log as logging
from oslo_utils import excutils
from sqlalchemy.orm import exc as orm_exception

from neutron.common import log
from neutron.db import models_v2
from neutron.i18n import _LE, _LI
from novaclient import exceptions as nova_exc
from novaclient.v1_1 import client as novaclient

from networking_brocade.vyatta.common import config
from networking_brocade.vyatta.common import exceptions as v_exc
from networking_brocade.vyatta.vrouter import client as vyatta_client


LOG = logging.getLogger(__name__)


class VyattaVRouterDriver(object):

    def __init__(self):
        self._vrouter_instance_map = {}
        self._management_network_id = config.VROUTER.management_network_id

        self._nova_client = novaclient.Client(
            config.VROUTER.tenant_admin_name,
            config.VROUTER.tenant_admin_password,
            auth_url=config.CONF.nova_admin_auth_url,
            service_type="compute",
            tenant_id=config.VROUTER.tenant_id)

    def create_router(self, context):
        LOG.debug("Vyatta vRouter Driver::Create Router")
        # Launch the vRouter VM. Method takes care of vRouter VM cleanup
        # when we encounter nova spawn or vRouter boot issues
        router = self._launch_routerVM(context)
        return router.id

    def init_router(self, context, router):
        LOG.debug("Vyatta vRouter Driver::Initialize router")
        try:
            vrouter_api = self._get_router_api(context, router['id'])
            vrouter_api.init_router(router.get('name', 'vyatta-router'),
                                    router.get('admin_state_up', False))
        except (v_exc.InvalidVRouterInstance,
                v_exc.InvalidInstanceConfiguration,
                v_exc.VRouterConnectFailure,
                v_exc.VRouterOperationError) as ex:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("Vyatta vRouter Driver::Initialize"
                              " router failed. Exception %s"), ex)
                self._delete_routerVM(context, router['id'])

    def delete_router(self, context, router_id):
        LOG.debug("Vyatta vRouter Driver::Deinitialize router")
        try:
            vrouter_api = self._get_router_api(context, router_id)
            vrouter_api.disconnect()
        finally:
            self._delete_routerVM(context, router_id)

    def attach_interface(self, context, router_id, port_id):
        LOG.debug("Vyatta vRouter Driver::Attach interface")
        router = self._nova_client.servers.get(router_id)
        router.interface_attach(port_id, None, None)

    def detach_interface(self, context, router_id, port_id):
        LOG.debug("Vyatta vRouter Driver::Deattach interface")
        router = self._nova_client.servers.get(router_id)
        router.interface_detach(port_id)

    def configure_interface(self, context, router_id, interface_infos):
        LOG.debug("Vyatta vRouter Driver::Configure interface")
        vrouter_api = self._get_router_api(context, router_id)
        for interface_info in interface_infos:
            vrouter_api.add_interface_to_router(interface_info)

    @log.log
    def update_interface(self, context, router_id, interface_info):
        vrouter_api = self._get_router_api(context, router_id)
        vrouter_api.update_interface(interface_info)

    def deconfigure_interface(self, context, router_id, interface_infos):
        LOG.debug("Vyatta vRouter Driver::Deconfigure interface")
        vrouter_api = self._get_router_api(context, router_id)
        for interface_info in interface_infos:
            vrouter_api.remove_interface_from_router(interface_info)

    def configure_gateway(self, context, router_id, interface_infos):
        LOG.debug("Vyatta vRouter Driver::Configure gateway")
        if len(interface_infos) != 1:
            raise v_exc.InvalidParameter(
                cause=_("Only one external gateway interface expected. "
                        "Given interfaces = %s") % len(interface_infos))
        vrouter_api = self._get_router_api(context, router_id)
        vrouter_api.update_router(external_gateway_info=interface_infos[0])

    def clear_gateway(self, context, router_id):
        LOG.debug("Vyatta vRouter Driver::Clear gateway")
        vrouter_api = self._get_router_api(context, router_id)
        vrouter_api.update_router(external_gateway_info=None)

    def assign_floating_ip(self, context, router_id, floating_ip, fixed_ip):
        LOG.debug("Vyatta vRouter Driver::Assign Floating IP")
        vrouter_api = self._get_router_api(context, router_id)
        vrouter_api.assign_floating_ip(floating_ip, fixed_ip)

    def unassign_floating_ip(self, context, router_id, floating_ip, fixed_ip):
        LOG.debug("Vyatta vRouter Driver::Unassign Floating IP")
        vrouter_api = self._get_router_api(context, router_id)
        vrouter_api.unassign_floating_ip(floating_ip, fixed_ip)

    def update_static_routes(self, context, router_id, route_add, route_del):
        LOG.debug('Vyatta vRouter Driver::Update static routes')
        vrouter_api = self._get_router_api(context, router_id)
        vrouter_api.update_static_routes(route_add, route_del)

    def _launch_routerVM(self, context):
        LOG.debug("Vyatta vRouter Driver::Launch router")
        router_name = 'vrouter_{0}'.format(os.urandom(6).encode('hex'))
        LOG.info(
            _LI("Vyatta vRouter Driver::Creating the vRouter instance %s"),
            router_name)

        try:
            router = self._nova_client.servers.create(
                router_name, config.VROUTER.image_id,
                config.VROUTER.flavor,
                nics=[{'net-id': self._management_network_id}])
        except (nova_exc.UnsupportedVersion, nova_exc.CommandError,
                nova_exc.AuthorizationFailure, nova_exc.NoUniqueMatch,
                nova_exc.AuthSystemNotFound, nova_exc.NoTokenLookupException,
                nova_exc.EndpointNotFound, nova_exc.AmbiguousEndpoints,
                nova_exc.ConnectionRefused, nova_exc.ClientException,
                Exception):
            with excutils.save_and_reraise_exception():
                LOG.error(
                    _LE("Vyatta vRouter Driver::Create server %s failed"),
                    router_name)
                raise v_exc.InstanceSpawnError()

        LOG.info(_LI("Vyatta vRouter Driver::Waiting for the vRouter "
                     "instance %s to start"), router_name)

        def _router_spawn():
            while True:
                try:
                    instance = self._nova_client.servers.get(router.id)
                except (nova_exc.UnsupportedVersion, nova_exc.CommandError,
                        nova_exc.AuthorizationFailure, nova_exc.NoUniqueMatch,
                        nova_exc.AuthSystemNotFound,
                        nova_exc.NoTokenLookupException,
                        nova_exc.EndpointNotFound, nova_exc.AmbiguousEndpoints,
                        nova_exc.ConnectionRefused, nova_exc.ClientException,
                        Exception):
                    yield config.VROUTER.nova_poll_interval
                    continue
                LOG.debug("Vyatta vRouter Driver::vRouter instance {0} "
                          "Spawn Status: {1}".format(router_name,
                                                     instance.status))
                if instance.status not in ('ACTIVE', 'ERROR'):
                    yield config.VROUTER.nova_poll_interval
                elif instance.status == 'ERROR':
                    raise v_exc.InstanceSpawnError()
                else:
                    break

        try:
            # Wait for Nova to spawn VM instance
            self._wait(_router_spawn,
                       timeout=config.VROUTER.nova_spawn_timeout)
        except (v_exc.InstanceSpawnError, v_exc.WaitTimeoutError) as ex:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE(
                    "Vyatta vRouter Driver::vRouter {0} spawn issue. "
                    "Exception {1}").format(router_name, ex))
                self._delete_routerVM(context, router.id)

        try:
            ifs = router.interface_list()
            if len(ifs) != 1:
                raise v_exc.InvalidParameter(
                    cause=_("Management interface expected "
                            "in router: %s") % router.id)
        except (nova_exc.UnsupportedVersion, nova_exc.CommandError,
                nova_exc.AuthorizationFailure, nova_exc.NoUniqueMatch,
                nova_exc.AuthSystemNotFound,
                nova_exc.NoTokenLookupException,
                nova_exc.EndpointNotFound, nova_exc.AmbiguousEndpoints,
                nova_exc.ConnectionRefused, nova_exc.ClientException,
                v_exc.InvalidParameter, Exception):
            with excutils.save_and_reraise_exception():
                self._delete_routerVM(context, router.id)

        def _router_boot():
            router_api = None
            while router_api is None:
                try:
                    router_api = self._get_router_api(context, router.id)
                except (v_exc.VRouterConnectFailure,
                        v_exc.VRouterOperationError):
                    yield config.VROUTER.vrouter_poll_interval
                    continue
                if router_api is not None:
                    break

        LOG.info(_LI("Vyatta vRouter Driver::Waiting for the vRouter {0} "
                     "to boot.").format(router_name))
        try:
            # Now wait for router to boot
            self._wait(_router_boot,
                       timeout=config.VROUTER.vrouter_boot_timeout)
        except (v_exc.WaitTimeoutError,
                v_exc.VRouterConnectFailure,
                v_exc.VRouterOperationError,
                v_exc.InvalidVRouterInstance,
                v_exc.InvalidInstanceConfiguration) as ex:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("Vyatta vRouter Driver::vRouter {0} boot issue. "
                              "Exception: {1}").format(router_name, ex))
                self._delete_routerVM(context, router.id)

        LOG.info(_LI("Vyatta vRouter Driver::vRouter instance %s is ready"),
                 router_name)

        return router

    def _wait(self, query_fn, timeout=0):
        end = time.time() + timeout
        for interval in query_fn():
            greenthread.sleep(interval)
            if timeout > 0 and time.time() >= end:
                raise v_exc.WaitTimeoutError()

    def _delete_routerVM(self, context, router_id):
        LOG.info(
            _LI("Vyatta vRouter Driver::Deleting the vRouter VM instance %s"),
            router_id)
        self._vrouter_instance_map.pop(router_id, None)

        try:
            self._nova_client.servers.delete(router_id)
        except (nova_exc.UnsupportedVersion, nova_exc.CommandError,
                nova_exc.AuthorizationFailure, nova_exc.NoUniqueMatch,
                nova_exc.AuthSystemNotFound,
                nova_exc.NoTokenLookupException,
                nova_exc.EndpointNotFound, nova_exc.AmbiguousEndpoints,
                nova_exc.ConnectionRefused, nova_exc.ClientException,
                Exception):
            with excutils.save_and_reraise_exception:
                LOG.error(
                    _LE("Vyatta vRouter Driver::Failed to delete the vRouter"
                        " VM instance %s"),
                    router_id)

    def _get_router_api(self, context, router_id):
        LOG.debug("Vyatta vRouter Driver::Get router driver")

        try:
            return self._vrouter_instance_map[router_id]
        except KeyError:

            try:
                query = context.session.query(models_v2.Network)
                network = query.filter_by(
                    id=self._management_network_id).one()
            except orm_exception.NoResultFound as ex:
                with excutils.save_and_reraise_exception():
                    LOG.error(_LE("Unable to find Vyatta vRouter "
                                  "management network %s"), ex)
                    raise v_exc.InvalidInstanceConfiguration(
                        cause="Unable to find management network")

            try:
                vrouter_instance = self._nova_client.servers.get(router_id)
            except nova_exc.ClientException as ex:
                with excutils.save_and_reraise_exception():
                    LOG.error(_LE(
                        "Unable to find Vyatta vRouter instance {0}. "
                        "Exception {1}").format(router_id, ex))
                    raise v_exc.InvalidVRouterInstance(router_id=router_id)

            LOG.debug("Vyatta vRouter Management network: %s",
                      network['name'])
            address_map = vrouter_instance.addresses[network['name']]
            if address_map is None:
                raise v_exc.InvalidVRouterInstance(router_id=router_id)
            address = address_map[0]["addr"]

            # Initialize vRouter API
            try:
                vrouter_api = vyatta_client.VRouterRestAPIClient()
                vrouter_api.connect(address)
            except Exception as ex:
                with excutils.save_and_reraise_exception():
                    LOG.error(_LE("Vyatta vRouter Driver: vRouter {0} "
                              "Connection exception {1}").format(address, ex))

            self._vrouter_instance_map[router_id] = vrouter_api
            return vrouter_api



# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.common import exceptions


class CorruptedSystemError(exceptions.NeutronException):
    message = _('System contain conflicts: %(description)s.')


class VRouterConnectFailure(exceptions.NeutronException):
    """Couldn't connect to instance."""
    message = _("Couldn't connect to Vyatta vRouter [%(ip_address)s].")


class VRouterOperationError(exceptions.NeutronException):
    """Internal Vyatta vRouter exception."""
    message = _("Internal Vyatta vRouter exception [%(ip_address)s]:"
                "%(reason)s.")


class InvalidVRouterInstance(exceptions.NeutronException):
    """Couldn't find the vrouter instance."""
    message = _("Couldn't find Vyatta vRouter instance %(router_id)s.")


class InvalidInstanceConfiguration(exceptions.NeutronException):
    """Invalid vRouter VM instance configuration."""
    message = _("Invalid Vyatta vRouter configuration: %(cause)s.")


class InvalidParameter(exceptions.NeutronException):
    """Invalid configuration parameter."""
    message = _("Invalid Parameter: %(cause)s.")


class InvalidResponseFormat(exceptions.NeutronException):
    message = _('Unparsable vRouter API response: %(details)s')


class WaitTimeoutError(exceptions.NeutronException):
    """Timeout error after waiting for Vyatta vRouter VM creation."""
    message = _("Timeout waiting for Vyatta vRouter instance creation.")


class InstanceSpawnError(exceptions.NeutronException):
    """vRouter VM instance spawning error."""
    message = _("Failed to spawn Vyatta vRouter VM instance.")


class InvalidL3AgentStateError(exceptions.NeutronException):
    message = _('Invalid L3 agent state: %(description)s')


class InvalidVPNServiceError(exceptions.NeutronException):
    message = _('Invalid or incomplete VPN service data')


class VPNResourceMappingNotFound(exceptions.NotFound):
    message = _('There is no VPN resource mapping for %(kind)s key=%(key)s')


class ResourceNotFound(exceptions.NotFound):
    message = _('Resource not found.')


class TableCellNotFound(exceptions.NotFound):
    message = _('There is no cell in vRouter status table.')


class DvrOrHaRouterNotSupported(exceptions.NeutronException):
    message = _('DVR or HA routers are not supported.')
